# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import shutil
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
from scipy.stats import beta, gamma
from tqdm import tqdm

from nemo.collections.asr.parts.preprocessing.perturb import AudioAugmentor
from nemo.collections.asr.parts.preprocessing.segment import AudioSegment
from nemo.collections.asr.parts.utils.manifest_utils import (
    get_ctm_line,
    read_manifest,
    write_ctm,
    write_manifest,
    write_text,
)
from nemo.collections.asr.parts.utils.speaker_utils import labels_to_rttmfile
from nemo.utils import logging


def get_cleaned_base_path(output_dir: str, overwrite_output: bool = True) -> str:
    """
    Delete output directory if it exists or throw warning.

    Args:
        output_dir (str): Path to output directory
        overwrite_output (bool): If True, delete output directory if it exists

    Returns:
        basepath (str): Path to base-path directory for writing output files
    """
    if os.path.isdir(output_dir) and os.listdir(output_dir):
        if overwrite_output:
            if os.path.exists(output_dir):
                shutil.rmtree(output_dir)
            os.mkdir(output_dir)
        else:
            raise Exception("Output directory is nonempty and overwrite_output = false")
    elif not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # only add root if paths are relative
    if not os.path.isabs(output_dir):
        ROOT = os.getcwd()
        basepath = os.path.join(ROOT, output_dir)
    else:
        basepath = output_dir
    return basepath


def binary_search_alignments(
    inds: List[int],
    max_audio_read_sec: float,
    min_alignment_count: int,
    alignments: List[float],
) -> int:
    """
    Binary search to find the index of the alignment that satisfies the maximum audio read duration,
    `max_audio_read_sec`. This is used to avoid reading the short audio files.
    NOTE: `offset_max` should be at least 1 to avoid feeding max=0 to random sampling function.

    Args:
        inds (list): List of indices to search from
        max_audio_read_sec (float): Maximum audio read duration
        min_alignment_count (int): Minimum number of alignments to read
        audio_manifest (dict): Dictionary containing the audio file's alignments

    Returns:
        offset_max (int) Index of the alignment that satisfies the maximum audio read duration
    """
    # Start from the left end (at index 0) and -1 * min_alignment_count for the right end
    left, right = 0, len(inds) - 1 - min_alignment_count
    while left < right:
        mid = left + (right - left) // 2
        dur_left = alignments[-1 * min_alignment_count] - alignments[inds[mid]]
        if dur_left < max_audio_read_sec:
            right = mid - 1
        elif dur_left > max_audio_read_sec:
            left = mid + 1
        else:
            break
    mid_out = left + (right - left) // 2
    # If mid_out is on the boundary, move it to the left.
    if alignments[-1 * min_alignment_count] - alignments[inds[mid_out]] < max_audio_read_sec:
        mid_out -= 1
    offset_max = max(mid_out, 1)
    return offset_max


def get_subset_of_audio_manifest(
    audio_manifest: dict,
    offset_index: int,
    max_audio_read_sec: float,
    min_alignment_count: int,
) -> dict:
    """
    Get a subset of `audio_manifest` for faster audio-file reading.

    Args:
        audio_manifest (dict): Audio manifest dictionary.
            keys: 'offset', 'duration', 'alignments', 'words'
        offset_index (int): Index of the offset.
        max_audio_read_sec (float): Maximum audio read duration.
        min_alignment_count (int): Minimum number of alignments to read.

    Returns:
        audio_manifest (dict): Subset of `audio_manifest` is returned for `words` and `alignments` keys.
    """
    alignment_array = np.array(audio_manifest['alignments'])
    alignment_array_pr = np.array(alignment_array[offset_index:]) - alignment_array[offset_index]
    subset_alignments = alignment_array_pr[alignment_array_pr < max_audio_read_sec]
    if len(subset_alignments) < min_alignment_count:
        # Cases where the word next to the offset is longer than the max_audio_read_sec.
        logging.warning(
            f"subset_alignments of {audio_manifest['audio_filepath']} \n"
            f"has subset alignment length:{len(subset_alignments)} at offset_index:{offset_index}, "
            f"word:{audio_manifest['words'][offset_index:offset_index+min_alignment_count]}, "
            f"alignments:{alignment_array_pr[:min_alignment_count]} which is longer than _max_audio_read_sec:{[0, max_audio_read_sec]}."
            " Truncating the alignements."
        )
        # Attach the `_max_audio_read_sec` to the `subset_alignments` to truncate the alignment timestamp.
        subset_alignments = np.concatenate([subset_alignments, np.array([max_audio_read_sec])])
    audio_manifest['offset'], audio_manifest['duration'] = (
        alignment_array[offset_index],
        subset_alignments[-1] - subset_alignments[0],
    )
    audio_manifest['alignments'] = subset_alignments.tolist()
    audio_manifest['words'] = audio_manifest['words'][offset_index : offset_index + len(subset_alignments)]
    return audio_manifest


def read_audio_from_buffer(
    audio_manifest: dict,
    buffer_dict: dict,
    offset_index: int,
    device: torch.device,
    max_audio_read_sec: float = 2.5,
    min_alignment_count: int = 2,
    read_subset: bool = True,
) -> Tuple[torch.Tensor, int, dict]:
    """
    Read from the provided file path while maintaining a hash-table that saves loading time.
    Also, this function only reads a subset of the audio file if `read_subset` is True for faster audio-file reading.

    Args:
        audio_manifest (dict): Audio manifest dictionary.
            keys: 'audio_filepath', 'duration', 'alignments', 'words'
        buffer_dict (dict): Hash-table that saves loaded audio files.
        offset_index (int): Index of the offset for the audio file.
        device (torch.device): Device to load the audio file.
        max_audio_read_sec (float): Maximum audio read duration.
        min_alignment_count (int): Minimum number of alignments to read.
        read_subset (bool): If True, read a subset of the audio file.
                            To control the length of the audio file, use data_simulator.session_params.max_audio_read_sec.
                            Note that using large value (greater than 3~4 sec) for `max_audio_read_sec` will slow down the generation process.
                            If False, read the entire audio file.

    Returns:
        audio_file (torch.Tensor): Time-series audio data in a tensor.
        sr (int): Sample rate of the audio file.
        audio_manifest (dict): (modified) audio manifest dictionary.
    """
    audio_file_id = f"{audio_manifest['audio_filepath']}#{offset_index}"
    if audio_file_id in buffer_dict:
        audio_file, sr, audio_manifest = buffer_dict[audio_file_id]
    else:
        if read_subset:
            audio_manifest = get_subset_of_audio_manifest(
                audio_manifest=audio_manifest,
                offset_index=offset_index,
                max_audio_read_sec=max_audio_read_sec,
                min_alignment_count=min_alignment_count,
            )
            segment = AudioSegment.from_file(
                audio_file=audio_manifest['audio_filepath'],
                offset=audio_manifest['offset'],
                duration=audio_manifest['duration'],
            )
        else:
            segment = AudioSegment.from_file(audio_file=audio_manifest['audio_filepath'])
        audio_file, sr = torch.from_numpy(segment.samples).to(device), segment.sample_rate
        if read_subset and segment.duration < (audio_manifest['alignments'][-1] - audio_manifest['alignments'][0]):
            audio_manifest['alignments'][-1] = min(segment.duration, audio_manifest['alignments'][-1])
        if audio_file.ndim > 1:
            audio_file = torch.mean(audio_file, 1, False).to(device)
        buffer_dict[audio_file_id] = (audio_file, sr, audio_manifest)
    return audio_file, sr, audio_manifest


def perturb_audio(
    audio: torch.Tensor, sr: int, augmentor: Optional[AudioAugmentor] = None, device: Optional[torch.device] = None
) -> torch.Tensor:
    """
    Perturb the audio (segment or session) using audio augmentor.

    Args:
        audio (torch.Tensor): Time-series signal of the segment
        sr (int): Sample rate of the original audio file
        augmentor (AudioAugmentor): Audio augmentor to use
        device (torch.device): Device to load the audio file

    Returns:
        audio (torch.Tensor): Perturbed audio (time-series signal) of the segment
    """
    if augmentor is None:
        return audio
    device = device if device is not None else torch.device('cpu')
    if isinstance(audio, torch.Tensor):
        audio = audio.cpu().numpy()
    audio_segment = AudioSegment(audio, sample_rate=sr)
    augmentor.perturb(audio_segment)
    audio_segment = torch.from_numpy(audio_segment.samples).to(device)
    return audio_segment


def normalize_audio(array: torch.Tensor) -> torch.Tensor:
    """
    Normalize the audio signal to avoid clipping.

    Args:
        array (torch.Tensor): Time-series audio data in a tensor.

    Returns:
        (torch.Tensor): Normalized audio signal.
    """
    return array / (1.0 * torch.max(torch.abs(array)))


def get_power_of_audio_file(audio_file: str, end_audio_file: int, running_len_samples: int, device: torch.device):
    """
    Calculate the power of the audio signal.

    Args:
        audio_file (torch.Tensor): Time-series audio data in a tensor.
        end_audio_file (int): End index of the audio file.
        running_len_samples (int): Running length of the audio file.
        device (torch.device): Device to use.

    Returns:
        (float): Power of the audio signal.
    """
    return torch.mean(audio_file[: end_audio_file - running_len_samples] ** 2).to(device)


def get_scaled_audio_signal(
    audio_file: torch.Tensor,
    end_audio_file: int,
    running_len_samples: int,
    desired_avg_power_noise: float,
    device: torch.device,
):
    """
    Scale the audio signal to the desired average power.

    Args:
        audio_file (torch.Tensor): Time-series audio data in a tensor.
        end_audio_file (int): End index of the audio file.
        running_len_samples (int): Running length of the audio file.
        desired_avg_power_noise (float): Desired average power of the audio file.
        device (torch.device): Device to use.

    Returns:
        scaled_audio_file (torch.Tensor): Scaled audio signal.
    """
    pow_audio_file = get_power_of_audio_file(
        audio_file=audio_file, end_audio_file=end_audio_file, running_len_samples=running_len_samples, device=device
    )
    scaled_audio_file = audio_file[: end_audio_file - running_len_samples] * torch.sqrt(
        desired_avg_power_noise / pow_audio_file
    ).to(device)
    return scaled_audio_file


def get_desired_avg_power_noise(
    power_array: float,
    snr_min: float,
    snr_max: float,
    background_noise_snr: float,
):
    """
    Calculate the desired average power of the noise.

    Args:
        power_array (float): Power of the audio signal.
        snr_min (float): Minimum SNR.
        snr_max (float): Maximum SNR.
        background_noise_snr (float): SNR of the background noise.

    Returns:
        desired_avg_power_noise (float): Desired average power of the noise.
    """
    if (snr_min is not None) and (snr_max is not None) and (snr_min <= snr_max):
        desired_snr = np.random.uniform(snr_min, snr_max)
    else:
        desired_snr = background_noise_snr
    ratio = 10 ** (desired_snr / 20)
    desired_avg_power_noise = power_array / ratio
    return desired_avg_power_noise, desired_snr


def get_background_noise(
    len_array: int,
    power_array: float,
    noise_samples: list,
    audio_read_buffer_dict: dict,
    snr_min: float,
    snr_max: float,
    background_noise_snr: float,
    seed: int,
    device: torch.device,
    sr: float = 16000,
):
    """
    Augment with background noise (inserting ambient background noise up to the desired SNR for the full clip).

    Args:
        len_array (int): Length of background noise required.
        power_array (float): Power of the audio signal.
        noise_samples (list): List of noise samples.
        audio_read_buffer_dict (dict): Dictionary containing audio read buffer.
        snr_min (float): Minimum SNR.
        snr_max (float): Maximum SNR.
        background_noise_snr (float): SNR of the background noise.
        seed (int): Seed for random number generator.
        device (torch.device): Device to use.

    Returns:
        bg_array (tensor): Tensor containing background noise.
        desired_snr (float): Desired SNR for adding background noise.
    """
    np.random.seed(seed)
    bg_array = torch.zeros(len_array).to(device)
    desired_avg_power_noise, desired_snr = get_desired_avg_power_noise(
        power_array=power_array, snr_min=snr_min, snr_max=snr_max, background_noise_snr=background_noise_snr
    )
    running_len_samples = 0
    noise_segment_list = []

    last_mixed_cut_offset = 0
    file_id = np.random.randint(len(noise_samples))
    while running_len_samples < len_array:  # build background audio stream (the same length as the full file)
        audio_file, sr, audio_manifest = read_audio_from_buffer(
            audio_manifest=noise_samples[file_id],
            buffer_dict=audio_read_buffer_dict,
            offset_index=0,
            device=device,
            read_subset=False,
        )
        # noise_segment_list.append(
        noise_manifest_dict = copy.deepcopy(audio_manifest)
        noise_manifest_dict['duration'] = float(min(len(audio_file), len_array - running_len_samples - 1) / sr)
        noise_manifest_dict['offset'] = 0
        noise_manifest_dict['volume'] = 1.0
        noise_manifest_dict['mixed_cut_offset'] = last_mixed_cut_offset
        last_mixed_cut_offset += noise_manifest_dict['duration']

        noise_segment_list.append(noise_manifest_dict)

        if running_len_samples + len(audio_file) < len_array:
            end_audio_file = running_len_samples + len(audio_file)
        else:
            end_audio_file = len_array
        scaled_audio_file = get_scaled_audio_signal(
            audio_file=audio_file,
            end_audio_file=end_audio_file,
            running_len_samples=running_len_samples,
            desired_avg_power_noise=desired_avg_power_noise,
            device=device,
        )

        bg_array[running_len_samples:end_audio_file] = scaled_audio_file
        running_len_samples = end_audio_file

    return bg_array, desired_snr, noise_segment_list


def get_random_offset_index(
    audio_manifest: dict,
    audio_read_buffer_dict: dict,
    offset_min: int = 0,
    max_audio_read_sec: float = 2.5,
    min_alignment_count: int = 2,
) -> int:
    """
    Get an index for randomly accessing the silence in alignment timestamps.

    Args:
        audio_manifest (dict): Audio manifest dictionary.
            keys: 'audio_filepath', 'duration', 'alignments', 'words'
        audio_read_buffer_dict (dict): Dictionary containing audio read buffer.
        offset_min (int): Minimum offset index. (Default: 0)
        max_audio_read_sec (float): Maximum audio read duration in seconds. (Default: 2.5)
        min_alignment_count (int): Minimum number of alignment timestamps. (Default: 2)

    Returns:
        (int): Random offset index smaller than `offset_count`.
    """
    if len(audio_manifest['alignments']) <= min_alignment_count:
        raise ValueError(
            f"Audio file {audio_manifest['audio_filepath']} has less than {min_alignment_count} alignment timestamps."
        )
    index_file_id = f"{audio_manifest['audio_filepath']}#index"

    # Avoid multiple indexings of the same audio file by using a hash-table.
    if index_file_id in audio_read_buffer_dict:
        (sil_inds, offset_max) = audio_read_buffer_dict[index_file_id]
    else:
        # Find all silence indices
        sil_inds = np.where((np.array(audio_manifest['words']) == '') == True)[0]
        if audio_manifest['alignments'][-1] - audio_manifest['alignments'][0] < max_audio_read_sec:
            # The total duration is already short, therefore skip range search.
            offset_max = 1
        else:
            # Find the range that satisfies `max_audio_read_sec` duration.
            offset_max = binary_search_alignments(
                inds=sil_inds,
                max_audio_read_sec=max_audio_read_sec,
                min_alignment_count=min_alignment_count,
                alignments=audio_manifest['alignments'],
            )

        audio_read_buffer_dict[index_file_id] = (sil_inds, offset_max)

    # If the audio file is shorter than the max_audio_read_sec, then we don't need to read a subset of the audio file.
    if (
        len(sil_inds) <= min_alignment_count
        or (audio_manifest['alignments'][-1] - audio_manifest['alignments'][0]) < max_audio_read_sec
    ):
        return offset_min
    else:
        offset_index = np.random.randint(offset_min, offset_max)
        return sil_inds[offset_index]


def get_speaker_ids(sess_idx: int, speaker_samples: dict, permutated_speaker_inds: list) -> List[str]:
    """
    Randomly select speaker IDs from the loaded manifest file.

    Args:
        sess_idx (int): Session index in integer.
        speaker_samples (dict): Dictionary mapping speaker ID to their list of samples.
        permutated_speaker_inds (list): List of permutated speaker indices.

    Returns:
        speaker_ids (list): List of speaker IDs
    """
    all_speaker_ids = list(speaker_samples.keys())
    # Measure the length of permutated_speaker_inds and mod the sess_idx number so that
    # sess_idx is always less than the length of permutated_speaker_inds
    sess_idx_circular = sess_idx % permutated_speaker_inds.shape[0]
    idx_list = permutated_speaker_inds[sess_idx_circular, :]
    speaker_ids = [all_speaker_ids[i] for i in idx_list]
    return speaker_ids


def build_speaker_samples_map(manifest: dict, tqdm_bar: bool = False) -> dict:
    """
    Build a dictionary for mapping speaker ID to their list of samples

    Returns:
        speaker_samples (Dict[list]):
            Dictionary mapping speaker ID to their list of samples
    """
    speaker_samples = defaultdict(list)
    # logging.info("Building speaker to samples map...")
    for sample in tqdm(manifest, total=len(manifest), disable=not tqdm_bar):
        # for sample in manifest:
        speaker_id = sample['speaker_id']
        speaker_samples[speaker_id].append(sample)
    return speaker_samples


def read_noise_manifest(add_bg: bool, background_manifest: str):
    """
    Read the noise manifest file and sample the noise manifest.

    Args:
        add_bg (bool): Whether to add background noise.
        background_manifest (str): Path to the background noise manifest file.

    Returns:
        noise_manifest (list): List of the entire noise source samples.
    """
    noise_manifest = []
    if add_bg is True:
        if background_manifest is not None:
            background_manifest_list = background_manifest
            if isinstance(background_manifest_list, str):
                background_manifest_list = [background_manifest_list]
            for background_manifest in background_manifest_list:
                if os.path.exists(background_manifest):
                    noise_manifest += read_manifest(background_manifest)
                else:
                    raise FileNotFoundError(f"Noise manifest file: {background_manifest} file not found.")
        else:
            raise FileNotFoundError(
                f"Noise manifest file is {background_manifest}. Please provide a valid noise manifest file/list if add_bg=True."
            )
    return noise_manifest


def read_rir_manifest(rir_manifest: str):
    """
    Read the rir manifest file and sample the rir manifest.
    """

    rir_manifest_list = [rir_manifest]
    rir_loaded_list = []
    for manifest_file in rir_manifest_list:
        if os.path.exists(manifest_file):
            rir_loaded_list.extend(read_manifest(manifest_file))

    return rir_loaded_list


def get_speaker_samples(speaker_ids: List[str], speaker_samples: dict) -> Dict[str, list]:
    """
    Get a list of the samples for each of the specified speakers.

    Args:
        speaker_ids (list): LibriSpeech speaker IDs for each speaker in the current session.
        speaker_samples (dict): Dictionary mapping speaker ID to their list of samples.

    Returns:
        speaker_wav_align_map (dict): Dictionary containing speaker IDs and their corresponding wav filepath and alignments.
    """
    speaker_wav_align_map = defaultdict(list)
    for sid in speaker_ids:
        speaker_wav_align_map[sid] = speaker_samples[sid]
    return speaker_wav_align_map


def add_silence_to_alignments(audio_manifest: dict):
    """
    Add silence to the beginning of the alignments and words.

    Args:
        audio_manifest (dict): Audio manifest dictionary.
            keys: 'audio_filepath', 'duration', 'alignments', 'words'

    Returns:
        audio_manifest (dict): Audio manifest dictionary with silence added to the beginning.
    """
    if type(audio_manifest['words'][0]) == str and len(audio_manifest['words'][0]) > 0:
        audio_manifest['words'].insert(0, "")
        audio_manifest['alignments'].insert(0, 0.0)
    return audio_manifest


def load_speaker_sample(
    speaker_wav_align_map: List[dict],
    speaker_ids: List[str],
    speaker_turn: int,
    min_alignment_count: int,
) -> str:
    """
    Load a sample for the selected speaker ID.
    The first alignment and word must be silence that determines the start of the alignments.

    Args:
        speaker_wav_align_map (dict): Dictionary containing speaker IDs and their corresponding wav filepath and alignments.
        speaker_ids (list): LibriSpeech speaker IDs for each speaker in the current session.
        speaker_turn (int): Current speaker turn.
        output_precision (int): Precision of the output alignments in integer.
        min_alignment_count (int): Minimum number of alignments in the audio file.

    Returns:
        audio_manifest (dict): Audio manifest dictionary containing the wav filepath, words and alignments.
    """
    speaker_id = speaker_ids[speaker_turn]
    file_id = np.random.randint(0, max(len(speaker_wav_align_map[str(speaker_id)]) - 1, 1))
    audio_manifest = speaker_wav_align_map[str(speaker_id)][file_id]

    # Check if the alignment file has at least 2 words.
    if len(audio_manifest['alignments']) < min_alignment_count:
        raise ValueError(
            f"Alignment file {audio_manifest['audio_filepath']} has an inappropriate length of {len(audio_manifest['alignments'])} < 2."
        )

    # Check whether the first word is silence and insert a silence token if the first token is not silence.
    if audio_manifest['words'][0] != "":
        audio_manifest = add_silence_to_alignments(audio_manifest)

    audio_manifest = copy.deepcopy(audio_manifest)
    return audio_manifest


def get_split_points_in_alignments(
    words: List[str],
    alignments: List[float],
    split_buffer: float,
    sr: int,
    sentence_audio_len: int,
    new_start: float = 0,
):
    """
    Collect split points in the alignment based on silence.
    Silence is defined as a blank symbol between two words that is longer than 2 * split_buffer.

    Args:
        words (List[str]): List of words in the sentence.
        alignments (List[float]): List of alignment timestamps in the sentence.
        split_buffer (float): Buffer length in seconds.
        sr (int): Sample rate of the audio.
        sentence_audio_len (int): Length of the sentence audio in samples.
        new_start (float): Start of the sentence audio in seconds.

    Returns:
        splits (List[List[int]]): List of integer split points in the sentence audio.
    """
    splits = []
    for i in range(len(words)):
        if words[i] == "" and i != 0 and i != len(words) - 1:
            # if words[i] == "<sil>" and i != 0 and i != len(words) - 1:
            silence_length = alignments[i] - alignments[i - 1]
            if silence_length > 2 * split_buffer:  # split utterance on silence
                new_end = alignments[i - 1] + split_buffer
                splits.append(
                    [
                        int(new_start * sr),
                        int(new_end * sr),
                    ]
                )
                new_start = alignments[i] - split_buffer
    # The last split point should be added
    splits.append([int(new_start * sr), sentence_audio_len])
    return splits


def per_speaker_normalize(
    sentence_audio: torch.Tensor, splits: List[List[int]], speaker_turn: int, volume: List[float], device: torch.device
) -> torch.Tensor:
    """
    Normalize time-series audio signal per speaker.

    Args:
        sentence_audio (torch.Tensor): Time-series audio signal.
        splits (List[List[int]]): List of integer split points in the sentence audio.
        speaker_turn (int): Speaker ID of the current speaker.
        volume (List[float]): List of volume levels for each speaker.
        device (torch.device): Device to use for computations.

    Returns:
        sentence_audio (torch.Tensor): Normalized time-series audio signal.
    """
    split_length = torch.tensor(0).to(device).double()
    split_sum = torch.tensor(0).to(device).double()
    for split in splits:
        split_length += len(sentence_audio[split[0] : split[1]])
        split_sum += torch.sum(sentence_audio[split[0] : split[1]] ** 2)
    average_rms = torch.sqrt(split_sum * 1.0 / split_length)
    sentence_audio = sentence_audio / (1.0 * average_rms) * volume[speaker_turn]
    return sentence_audio


class DataAnnotator(object):
    """
    Class containing the functions that create RTTM, CTM, JSON files.

    Arguments in config:

    data_simulator:
        session_config:
            num_speakers (int): Number of unique speakers per multispeaker audio session
            session_params:
            split_buffer (float): Split RTTM labels if greater than twice this amount of silence (to avoid long gaps between
                                    utterances as being labelled as speech)
        outputs:
            output_dir (str): Output directory for audio sessions and corresponding label files
            output_filename (str): Output filename for the wav and RTTM files
            overwrite_output (bool): If true, delete the output directory if it exists
            output_precision (int): Number of decimal places in output files
    """

    def __init__(self, cfg):
        """
        Args:
            cfg: OmegaConf configuration loaded from yaml file.
        """
        self._params = cfg
        self._files = {}
        self._init_file_write()
        self._init_filelist_lists()

    def _init_file_write(self):
        """
        Initialize file writing arguments
        """
        self._file_base_str = "synthetic"
        self._file_types = ["wav", "rttm", "json", "noise", "ctm", "txt", "meta"]
        self._annotation_types = ["rttm", "json", "ctm"]

    def _init_filelist_lists(self):
        """
        Initialize lists to store the filelists for each file type
        """
        self.annote_lists = {}
        for file_type in self._file_types:
            self.annote_lists[f"{file_type}_list"] = []

    def init_annotation_lists(self):
        """
        Initialize lists to store the annotations for each file type
        """
        for file_type in self._file_types:
            self.annote_lists[file_type] = []

    def create_new_rttm_entry(
        self,
        words: List[str],
        alignments: List[float],
        start: int,
        end: int,
        speaker_id: int,
        add_split_buffer: bool = False,
    ) -> List[str]:
        """
        Create new RTTM entries (to write to output rttm file)

        Args:
            words (list): List of words in the current audio file.
            alignments (list): List of alignments (timestamps) for the current audio file.
            start (int): Current start of the audio file being inserted.
            end (int): End of the audio file being inserted.
            speaker_id (int): LibriSpeech speaker ID for the current entry.

        Returns:
            rttm_list (list): List of rttm entries
        """
        rttm_list = []
        new_start = start
        # look for split locations
        for i in range(len(words)):
            if words[i] == "" and i != 0 and i != len(words) - 1:
                silence_length = alignments[i] - alignments[i - 1]
                if (
                    silence_length > 2 * self._params.data_simulator.session_params.split_buffer
                ):  # split utterance on silence
                    new_end = start + alignments[i - 1]

                    # new_end = start + alignments[i - 1] + self._params.data_simulator.session_params.split_buffer

                    # import ipdb; ipdb.set_trace()
                    # if add_split_buffer:  # add split buffer if specified in config
                    #     new_end += self._params.data_simulator.session_params.split_buffer
                    t_stt = round(float(new_start), self._params.data_simulator.outputs.output_precision)
                    t_end = round(float(new_end), self._params.data_simulator.outputs.output_precision)
                    rttm_list.append(f"{t_stt} {t_end} {speaker_id}")
                    new_start = start + alignments[i]
                    # new_start = start + alignments[i] - self._params.data_simulator.session_params.split_buffer

        t_stt = round(float(new_start), self._params.data_simulator.outputs.output_precision)
        t_end = round(float(end), self._params.data_simulator.outputs.output_precision)
        rttm_list.append(f"{t_stt} {t_end} {speaker_id}")
        return rttm_list

    def create_new_json_entry(
        self,
        text: List[str],
        wav_filename: str,
        start: float,
        length: float,
        speaker_id: int,
        rttm_filepath: str,
        ctm_filepath: str,
    ) -> dict:
        """
        Create new JSON entries (to write to output json file).

        Args:
            text (list): string of text for the current entry.
            wav_filename (str): Filename of the wav file.
            start (float): Start time of the current entry.
            length (float): Length of the current entry.
            speaker_id (int): speaker ID for the current entry.
            rttm_filepath (str): Path to the RTTM file.
            ctm_filepath (str): Path to the CTM file.

        Returns:
            meta (dict): JSON entry dictionary.
        """
        start = round(float(start), self._params.data_simulator.outputs.output_precision)
        length = round(float(length), self._params.data_simulator.outputs.output_precision)
        meta = {
            "audio_filepath": wav_filename,
            "offset": start,
            "duration": length,
            "label": speaker_id,
            "text": text,
            "num_speakers": self._params.data_simulator.session_config.num_speakers,
            "rttm_filepath": rttm_filepath,
            "ctm_filepath": ctm_filepath,
            "uem_filepath": None,
        }
        return meta

    def create_ctm_entry_from_segment_list(
        self, source_segment_list, session_name: str, speaker_id: int, start: int
    ) -> List[str]:
        """
        Create new CTM entry (to write to output ctm file)

        Args:
            words (list): List of words in the current audio file.
            alignments (list): List of alignments (timestamps) for the current audio file.
            session_name (str): Current session name.
            speaker_id (int): LibriSpeech speaker ID for the current entry.
            start (int): Current start of the audio file being inserted.

        Returns:
            arr (list): List of ctm entries
        """
        arr = []
        start = float(round(start, self._params.data_simulator.outputs.output_precision))

        for seg_dict in source_segment_list:
            words = seg_dict["words"]
            alignments = seg_dict["alignments"]
            start_offset = seg_dict["mixed_cut_offset"]
            alignment_offset = alignments[0]
            for i in range(len(words)):
                word = words[i]
                if (
                    word != ""
                ):  # note that using the current alignments the first word is always empty, so there is no error from indexing the array with i-1
                    # prev_align = 0 if i == 0 else alignments[i - 1]
                    # align1 = round(float(prev_align + start), self._params.data_simulator.outputs.output_precision)
                    align1 = round(
                        float(start_offset + alignments[i] - alignment_offset),
                        self._params.data_simulator.outputs.output_precision,
                    )
                    align2 = round(
                        float(start_offset + alignments[i + 1] - alignment_offset - align1),
                        self._params.data_simulator.outputs.output_precision,
                    )
                    text = get_ctm_line(
                        source=session_name,
                        channel=1,
                        start_time=align1,
                        duration=align2,
                        token=word,
                        conf=None,
                        type_of_token='lex',
                        speaker=speaker_id,
                    )
                    arr.append((align1, text))
        return arr

    def create_new_ctm_entry(
        self,
        words: List[str],
        alignments: List[float],
        session_name: str,
        speaker_id: int,
        start: int,
    ) -> List[str]:
        """
        Create new CTM entry (to write to output ctm file)

        Args:
            words (list): List of words in the current audio file.
            alignments (list): List of alignments (timestamps) for the current audio file.
            session_name (str): Current session name.
            speaker_id (int): LibriSpeech speaker ID for the current entry.
            start (int): Current start of the audio file being inserted.

        Returns:
            arr (list): List of ctm entries
        """
        arr, word_and_ts_list = [], []
        start = float(round(start, self._params.data_simulator.outputs.output_precision))

        for i in range(len(words)):
            word = words[i]
            if (
                word != ""
            ):  # note that using the current alignments the first word is always empty, so there is no error from indexing the array with i-1
                prev_align = 0 if i == 0 else alignments[i - 1]
                align1 = round(float(prev_align + start), self._params.data_simulator.outputs.output_precision)
                align2 = round(float(alignments[i] - prev_align), self._params.data_simulator.outputs.output_precision)
                end_time = round(align1 + align2, self._params.data_simulator.outputs.output_precision)
                text = get_ctm_line(
                    source=session_name,
                    channel=1,
                    start_time=align1,
                    duration=align2,
                    token=word,
                    conf=None,
                    type_of_token='lex',
                    speaker=speaker_id,
                    output_precision=self._params.data_simulator.outputs.output_precision,
                )
                word_and_ts_list.append((word, align1, end_time))
                arr.append((align1, text))
        return arr, word_and_ts_list

    def add_to_filename_lists(self, basepath: str, filename: str):
        """
        Add the current filename to the list of filenames for each file type.

        Args:
            basepath (str): Basepath for output files.
            filename (str): Base filename for all output files.
        """
        full_base_filepath = os.path.join(basepath, filename)
        for file_type in self._file_types:
            self.annote_lists[f"{file_type}_list"].append(f"{full_base_filepath}.{file_type}")

    def write_filelist_files(self, basepath):
        """
        Write all filelist files.

        Args:
            basepath (str): Basepath for output files.
        """
        for file_type in self._file_types:
            with open(f"{basepath}/{self._file_base_str}_{file_type}.list", "w") as list_file:
                list_file.write("\n".join(self.annote_lists[f"{file_type}_list"]))
            list_file.close()

    def write_annotation_files(self, basepath: str, filename: str, meta_data: dict):
        """
        Write all annotation files: RTTM, JSON, CTM, TXT, and META.

        Args:
            basepath (str): Basepath for output files.
            filename (str): Base filename for all output files.
            meta_data (dict): Metadata for the current session.
            rttm_list (list): List of RTTM entries.
            json_list (list): List of JSON entries.
            ctm_list (list): List of CTM entries.
        """
        labels_to_rttmfile(self.annote_lists['rttm'], filename, self._params.data_simulator.outputs.output_dir)
        write_manifest(os.path.join(basepath, filename + '.json'), self.annote_lists['json'])
        write_ctm(os.path.join(basepath, filename + '.ctm'), self.annote_lists['ctm'])
        write_text(os.path.join(basepath, filename + '.txt'), self.annote_lists['ctm'])
        write_manifest(os.path.join(basepath, filename + '.meta'), [meta_data])

    def write_annotation_rttm_and_ctm(self, basepath: str, filename: str):
        """
        Write all annotation files: RTTM, JSON, CTM, TXT, and META.

        Args:
            basepath (str): Basepath for output files.
            filename (str): Base filename for all output files.
            meta_data (dict): Metadata for the current session.
            rttm_list (list): List of RTTM entries.
            json_list (list): List of JSON entries.
            ctm_list (list): List of CTM entries.
        """
        labels_to_rttmfile(
            self.annote_lists['rttm'], os.path.join(basepath, filename), self._params.data_simulator.outputs.output_dir
        )
        write_ctm(os.path.join(basepath, filename + '.ctm'), self.annote_lists['ctm'])


class SpeechSampler(object):
    """
    Class for sampling speech samples for Multispeaker Audio Session Simulator

    Args:
        cfg: OmegaConf configuration loaded from yaml file.

    Variables for sampling speech:
        self.running_speech_len_samples (int): Running total of speech samples in the current audio session.
        self.running_silence_len_samples (int): Running total of silence samples in the current audio session.
        self.running_overlap_len_samples (int): Running total of overlap samples in the current audio session.

        self.sess_silence_mean (int) : Targeted mean number of silence samples in the current audio session.
        self.per_silence_min_len (int): Minimum number of silence samples in the silence segment.
        self.per_silence_max_len (int): Maximum number of silence samples in the silence segment.

        self.sess_overlap_mean (int): Targeted mean number of overlap samples in the current audio session.
        self.per_overlap_min_len (int): Minimum number of overlap samples in the overlap segment.
        self.per_overlap_max_len (int): Maximum number of overlap samples in the overlap segment.

    data_simulator:
        session_params:
            mean_silence (float): Mean proportion of silence to speaking time in the audio session. Should be in range [0, 1).
            mean_silence_var (float): Variance for mean silence in all audio sessions.
                                    This value should be 0 <= mean_silence_var < mean_silence * (1 - mean_silence).
            per_silence_var (float):  Variance for each silence in an audio session, set large values (e.g., 20) for de-correlation.
            per_silence_min (float): Minimum duration for each silence, default to 0.
            per_silence_max (float): Maximum duration for each silence, default to -1 for no maximum.

            mean_overlap (float): Mean proportion of overlap in the overall non-silence duration. Should be in range [0, 1) and
                                recommend [0, 0.15] range for accurate results.
            mean_overlap_var (float): Variance for mean overlap in all audio sessions.
                                    This value should be 0 <= mean_overlap_var < mean_overlap * (1 - mean_overlap).
            per_overlap_var (float): Variance for per overlap in each session, set large values to de-correlate silence lengths
                                    with the latest speech segment lengths
            per_overlap_min (float): Minimum per overlap duration in seconds
            per_overlap_max (float): Maximum per overlap duration in seconds, set -1 for no maximum
    """

    def __init__(self, cfg):
        """
        Args:
            cfg: OmegaConf configuration loaded from yaml file.
        """
        self._params = cfg

        self.running_speech_len_samples = 0
        self.running_silence_len_samples = 0
        self.running_overlap_len_samples = 0

        self.sess_silence_mean = None
        self.per_silence_min_len = 0
        self.per_silence_max_len = 0

        self.sess_overlap_mean = None
        self.per_overlap_min_len = 0
        self.per_overlap_max_len = 0

        self.mean_overlap = float(self._params.data_simulator.session_params.mean_overlap)
        self.mean_overlap_var = float(self._params.data_simulator.session_params.mean_overlap_var)

        self.mean_silence = float(self._params.data_simulator.session_params.mean_silence)
        self.mean_silence_var = float(self._params.data_simulator.session_params.mean_silence_var)

        self.per_silence_var = float(self._params.data_simulator.session_params.per_silence_var)
        self.per_overlap_var = float(self._params.data_simulator.session_params.per_overlap_var)

        self.num_noise_files = int(self._params.data_simulator.background_noise.num_noise_files)

    def _mean_var_to_a_and_b(self, mean: float, var: float) -> Tuple[float, float]:
        """
        Convert mean and variance to a and b parameters for beta distribution.

        Args:
            mean (float): Mean of the beta distribution.
            var (float): Variance of the beta distribution.

        Returns:
            Tuple[float, float]: a and b parameters for beta distribution.
        """
        a = mean**2 * (1 - mean) / var - mean
        b = mean * (1 - mean) ** 2 / var - (1 - mean)
        return a, b

    def _init_silence_params(self):
        """
        Initialize parameters for silence insertion in the current session.
        """
        self.running_speech_len_samples = 0
        self.running_silence_len_samples = 0

        self.per_silence_min_len = int(
            max(0, self._params.data_simulator.session_params.per_silence_min) * self._params.data_simulator.sr
        )
        if self._params.data_simulator.session_params.per_silence_max > 0:
            self.per_silence_max_len = int(
                self._params.data_simulator.session_params.per_silence_max * self._params.data_simulator.sr
            )
        else:
            self.per_silence_max_len = int(
                self._params.data_simulator.session_config.session_length * self._params.data_simulator.sr
            )

    def _init_overlap_params(self):
        """
        Initialize parameters for overlap insertion in the current session.
        """
        self.running_overlap_len_samples = 0

        self.per_overlap_min_len = int(
            max(0, self._params.data_simulator.session_params.per_overlap_min) * self._params.data_simulator.sr
        )
        if self._params.data_simulator.session_params.per_overlap_max > 0:
            self.per_overlap_max_len = int(
                self._params.data_simulator.session_params.per_overlap_max * self._params.data_simulator.sr
            )
        else:
            self.per_overlap_max_len = int(
                self._params.data_simulator.session_config.session_length * self._params.data_simulator.sr
            )

    def silence_vs_overlap_selector(self, running_len_samples: int, non_silence_len_samples: int) -> bool:
        """
        Compare the current silence ratio to the current overlap ratio. Switch to either silence or overlap mode according
        to the amount of the gap between current ratio and session mean in config.

        Args:
            running_len_samples (int): Length of the current session in samples.
            non_silence_len_samples (int): Length of the signal that is not silence in samples.

        Returns:
            add_overlap (bool): True if the current silence ratio is less than the current overlap ratio, False otherwise.
        """
        if running_len_samples > 0:
            self.current_silence_ratio = (running_len_samples - self.running_speech_len_samples) / running_len_samples
            self.current_overlap_ratio = self.running_overlap_len_samples / non_silence_len_samples
        else:
            self.current_silence_ratio, self.current_overlap_ratio = 0, 0

        # self.silence_discrepancy = max(0, self.sess_silence_mean - self.current_silence_ratio)
        # self.overlap_discrepancy = max(0, self.sess_overlap_mean - self.current_overlap_ratio)
        # threshold = self.silence_discrepancy / (self.overlap_discrepancy + self.silence_discrepancy + 1e-10)
        # add_overlap = np.random.rand() > threshold
        self.silence_discrepancy = self.current_silence_ratio - self.sess_silence_mean
        self.overlap_discrepancy = self.current_overlap_ratio - self.sess_overlap_mean
        add_overlap = bool(self.overlap_discrepancy < self.silence_discrepancy)
        return add_overlap

    def get_session_silence_mean(self):
        """
        Get the target mean silence for current session using re-parameterized Beta distribution.
        The following constraints are applied to make a > 0 and b > 0:

            0 < mean_silence < 1
            0 < mean_silence_var < mean_silence * (1 - mean_silence)

        Args:
            silence_mean (float):
                Target mean silence for the current session
        """
        self._init_silence_params()
        mean, var = self.mean_silence, self.mean_silence_var
        if var > 0:
            a, b = self._mean_var_to_a_and_b(mean, var)
            if a < 0 or b < 0:
                raise ValueError(
                    f"Beta(a, b), a = {a:.3f} and b = {b:.3f} should be both greater than 0. "
                    f"Invalid `mean_silence_var` value {var} for sampling from Beta distribution. "
                    f"`mean_silence_var` should be less than `mean_silence * (1 - mean_silence)`. "
                    f"Please check `mean_silence_var` and try again."
                )
            self.sess_silence_mean = beta(a, b).rvs()
        else:
            self.sess_silence_mean = mean
        return self.sess_silence_mean

    def get_session_overlap_mean(self):
        """
        Get the target mean overlap for current session using re-parameterized Beta distribution.
        The following constraints are applied to make a > 0 and b > 0:

            0 < mean_overlap < 1
            0 < mean_overlap_var < mean_overlap * (1 - mean_overlap)

        Returns:
            overlap_mean (float):
                Target mean overlap for the current session
        """
        self._init_overlap_params()
        mean, var = self.mean_overlap, self.mean_overlap_var
        if var > 0:
            a, b = self._mean_var_to_a_and_b(mean, var)
            if a < 0 or b < 0:
                raise ValueError(
                    f"Beta(a, b), a = {a:.3f} and b = {b:.3f} should be both greater than 0. "
                    f"Invalid `mean_overlap_var` value {var} for sampling from Beta distribution. "
                    f"`mean_overlap_var` should be less than `mean_overlap * (1 - mean_overlap)`. "
                    f"Please check `mean_overlap_var` and try again."
                )
            self.sess_overlap_mean = beta(a, b).rvs()
        else:
            self.sess_overlap_mean = mean
        return self.sess_overlap_mean

    def sample_from_silence_model(self, running_len_samples: int) -> int:
        """
        Sample from the silence model to determine the amount of silence to add between sentences.
        Gamma distribution is employed for modeling the highly skewed distribution of silence length distribution.
        When we add silence between sentences, we want to ensure that the proportion of silence meets the `sess_silence_mean`.
        Thus, [Session Silence Mean] = [Total Running Silence Time] / [Total Running Session Time] equation holds. We employ the following
        formula to determine the amount of silence to add, which is `silence_mean`:

            self.sess_silence_mean = (silence_mean + self.running_silence_len_samples) / (silence_mean + running_len_samples)

        The above equation is setting `silence_mean` to yield the desired silence ratio `self.sess_silence_mean`.
        We use the above `silence_mean` value to sample silence-length for each silence occurrence.

        Args:
            running_len_samples (int):
                Running length of the session (in terms of number of samples).
            session_len_samples (int):
                Targeted total session length (in terms of number of samples).

        Returns:
            silence_amount (int): Amount of silence to add between sentences (in terms of number of samples).
        """
        silence_mean = ((self.sess_silence_mean * running_len_samples) - self.running_silence_len_samples) / (
            1 - self.sess_silence_mean
        )
        silence_mean = max(self.per_silence_min_len, min(silence_mean, self.per_silence_max_len))
        if silence_mean > 0:
            self.per_silence_var = self._params.data_simulator.session_params.per_silence_var
            silence_amount = (
                int(gamma(a=(silence_mean**2) / self.per_silence_var, scale=self.per_silence_var / silence_mean).rvs())
                if self.per_silence_var > 0
                else int(silence_mean)
            )
            silence_amount = max(self.per_silence_min_len, min(silence_amount, self.per_silence_max_len))
        else:
            silence_amount = 0
        return silence_amount

    def sample_from_overlap_model(self, non_silence_len_samples: int):
        """
        Sample from the overlap model to determine the amount of overlap between segments.
        Gamma distribution is employed for modeling  the highly skewed distribution of overlap length distribution.
        When we add an overlap occurrence, we want to meet the desired overlap ratio defined by `self.sess_overlap_mean`.
        Thus, [Session Overlap Mean] = [Total Running Overlap Speech Time] / [Total Running Non-Silence Speech Time].
        Let `overlap_mean` be the desired overlap amount, then the mean and variance of the gamma distribution is given by:

            self.sess_overlap_mean = (overlap_mean + self.running_overlap_len_samples) / (non_silence_len_samples - overlap_mean)

        The above equation is setting `overlap_mean` to yield the desired overlap ratio `self.sess_overlap_mean`.
        We use the above `overlap_mean` value to sample overlap-length for each overlap occurrence.

        Args:
            non_silence_len_samples (int):
                The total amount of non-silence (speech) region regardless of overlap status

        Returns:
            desired_overlap_amount (int):
                Amount of overlap between segments (in terms of number of samples).
        """
        overlap_mean = ((self.sess_overlap_mean * non_silence_len_samples) - self.running_overlap_len_samples) / (
            1 + self.sess_overlap_mean
        )
        overlap_mean = max(self.per_overlap_min_len, min(max(0, overlap_mean), self.per_overlap_max_len))

        if overlap_mean > 0:
            desired_overlap_amount = (
                int(gamma(a=overlap_mean**2 / self.per_overlap_var, scale=self.per_overlap_var / overlap_mean).rvs())
                if self.per_overlap_var > 0
                else int(overlap_mean)
            )
            desired_overlap_amount = max(
                self.per_overlap_min_len, min(desired_overlap_amount, self.per_overlap_max_len)
            )
        else:
            desired_overlap_amount = 0
        return desired_overlap_amount

    def sample_noise_manifest(self, noise_manifest: dict) -> list:
        """
        Sample noise manifest to a specified count `num_noise_files` for the current simulated audio session.

        Args:
            noise_manifest (list):
                List of noise source samples to be sampled from.

        Returns:
            sampled_noise_manifest (list):
                List of noise samples to be used for the current session.
        """
        num_noise_files = min(len(noise_manifest), self.num_noise_files)
        sampled_noise_manifest = []
        if num_noise_files > 0:
            selected_noise_ids = np.random.choice(range(len(noise_manifest)), num_noise_files, replace=False)
            for k in selected_noise_ids:
                sampled_noise_manifest.append(noise_manifest[k])
        return sampled_noise_manifest



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import math
import os
import time
from collections import OrderedDict
from copy import deepcopy
from functools import wraps
from typing import Any, Dict, List, Optional, Tuple

import torch
from lhotse.dataset.collation import collate_matrices
from omegaconf import DictConfig

from nemo.collections.asr.data.audio_to_diar_label import extract_frame_info_from_rttm, get_frame_targets_from_rttm
from nemo.collections.asr.models.sortformer_diar_models import SortformerEncLabelModel
from nemo.collections.asr.modules.sortformer_modules import StreamingSortformerState
from nemo.collections.asr.parts.utils.diarization_utils import get_color_palette, print_sentences, write_txt
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.asr.parts.utils.speaker_utils import audio_rttm_map as get_audio_rttm_map
from nemo.collections.asr.parts.utils.speaker_utils import get_uniqname_from_filepath
from nemo.utils import logging


def measure_eta(func):
    """
    Measure the time taken to execute the function and print the ETA.

    Args:
        func (callable): The function to measure the ETA of.

    Returns:
        callable: The wrapped function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()  # Record the start time
        result = func(*args, **kwargs)  # Execute the function
        end_time = time.time()  # Record the end time
        eta = end_time - start_time  # Calculate the elapsed time
        logging.info(f"[ Step-{kwargs['step_num']} ] for '{func.__name__}': {eta:.4f} seconds")  # Print the ETA
        return result  # Return the original function's result

    return wrapper


def format_time(seconds: float) -> str:
    """
    Format the time in minutes and seconds.

    Args:
        seconds (float): The time in seconds.

    Returns:
        str: The time in minutes and seconds.
    """
    minutes = math.floor(seconds / 60)
    sec = seconds % 60
    return f"{minutes}:{sec:05.2f}"


def add_delay_for_real_time(
    cfg: Any,
    chunk_audio: torch.Tensor,
    session_start_time: float,
    feat_frame_count: int,
    loop_end_time: float,
    loop_start_time: float,
):
    """
    Add artificial delay for real-time mode by calculating the time difference between
    the current time and the session start time..

    Args:
        cfg (Any): The configuration object containing the parameters for the delay calculation.
        chunk_audio (torch.Tensor): The chunk audio tensor containing time-series audio data.
        session_start_time (float): The session start time in seconds.
        feat_frame_count (int): The number of features per second.
        loop_end_time (float): The loop end time in seconds.
        loop_start_time (float): The loop start time in seconds.
    """
    time_diff = max(0, (time.time() - session_start_time) - feat_frame_count * cfg.get("feat_len_sec", 0.01))
    eta_min_sec = format_time(time.time() - session_start_time)
    logging.info(
        f"[   REAL TIME MODE   ] min:sec - {eta_min_sec} "
        f"Time difference for real-time mode: {time_diff:.4f} seconds"
    )
    time.sleep(
        max(
            0,
            (chunk_audio.shape[-1] - cfg.get("discarded_frames", 8)) * cfg.get("feat_len_sec", 0.01)
            - (loop_end_time - loop_start_time)
            - time_diff * cfg.get("finetune_realtime_ratio", 0.01),
        )
    )


def write_seglst_file(seglst_dict_list: List[Dict[str, Any]], output_path: str):
    """
    Write a seglst file from the seglst dictionary list.

    Args:
        seglst_dict_list (List[Dict[str, Any]]): The list of seglst dictionaries.
            Example:
            [
                {
                    "session_id": "session_001",
                    "speaker": "speaker_1",
                    "words": "Write this to a SegLST file.",
                    "start_time": 12.34,
                    "end_time": 23.45,
                }, ...
            ]
        output_path (str): The path to the output file.
    """
    if len(seglst_dict_list) == 0:
        raise ValueError("seglst_dict_list is empty. No transcriptions were generated.")
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(json.dumps(seglst_dict_list, indent=4) + '\n')
    logging.info(f"Saved the transcriptions of the streaming inference in\n:{output_path}")


def get_multi_talker_samples_from_manifest(cfg, manifest_file: str, feat_per_sec: float, max_spks: int):
    """
    Get the multi-talker samples from the manifest file and save it to a list named 'samples'.
    Also, save the rttm mask matrix to a list named 'rttms_mask_mats'.

    Args:
        cfg (DictConfig): The configuration object.
        manifest_file (str): The path to the manifest file.
        feat_per_sec (float): The number of features per second.
        max_spks (int): The maximum number of speakers.

    Returns:
        samples (list): The list of samples.
        rttms_mask_mats (list): The list of rttm mask matrices.
    """
    samples, rttms_mask_mats = [], []
    with open(manifest_file, 'r', encoding='utf-8') as f:
        for line_num, line in enumerate(f):
            item = json.loads(line)
            if 'audio_filepath' not in item:
                raise KeyError(f"Line {line_num}: 'audio_filepath' missing")
            if 'duration' not in item:
                raise KeyError(f"Line {line_num}: 'duration' missing")
            samples.append(item)
            if cfg.get("spk_supervision", "diar") == "rttm":
                rttm_path = samples[-1]['rttm_filepath']
                if not rttm_path:
                    raise ValueError(f"Line {line_num}: rttm_filepath required when spk_supervision='rttm'")
                if not os.path.exists(rttm_path):
                    raise FileNotFoundError(f"Line {line_num}: RTTM file not found: {rttm_path}")

                with open(rttm_path, 'r', encoding='utf-8') as f:
                    rttm_lines = f.readlines()
                rttm_timestamps, _ = extract_frame_info_from_rttm(0, samples[-1]['duration'], rttm_lines)
                rttm_mat = get_frame_targets_from_rttm(
                    rttm_timestamps=rttm_timestamps,
                    offset=0,
                    duration=samples[-1]['duration'],
                    round_digits=3,
                    feat_per_sec=round(float(1 / feat_per_sec), 2),
                    max_spks=max_spks,
                )
                rttms_mask_mats.append(rttm_mat)
            samples[-1]['duration'] = None
            if 'offset' not in item:
                samples[-1]['offset'] = 0

    if len(rttms_mask_mats) > 0:
        rttms_mask_mats = collate_matrices(rttms_mask_mats)
    else:
        rttms_mask_mats = None
    return samples, rttms_mask_mats


def setup_diarization_model(cfg: DictConfig, map_location: Optional[str] = None) -> SortformerEncLabelModel:
    """Setup model from cfg and return diarization model and model name for next step"""
    if cfg.diar_model_path.endswith(".ckpt"):
        diar_model = SortformerEncLabelModel.load_from_checkpoint(
            checkpoint_path=cfg.diar_model_path, map_location=map_location, strict=False
        )
        model_name = os.path.splitext(os.path.basename(cfg.diar_model_path))[0]
    elif cfg.diar_model_path.endswith(".nemo"):
        diar_model = SortformerEncLabelModel.restore_from(restore_path=cfg.diar_model_path, map_location=map_location)
        model_name = os.path.splitext(os.path.basename(cfg.diar_model_path))[0]
    elif cfg.diar_pretrained_name.startswith("nvidia/"):
        diar_model = SortformerEncLabelModel.from_pretrained(cfg.diar_pretrained_name)
        model_name = os.path.splitext(os.path.basename(cfg.diar_pretrained_name))[0]
    else:
        raise ValueError("cfg.diar_model_path must end with.ckpt or.nemo!")
    return diar_model, model_name


def write_seglst(output_filepath: str, seglst_list: list) -> None:
    """
    Write the segmentation list to a file.

    Args:
        output_filepath (str): The path to the output file.
        seglst_list (list): The list of segmentation lists.
    """
    with open(output_filepath, "w", encoding="utf-8") as f:
        f.write(json.dumps(seglst_list, indent=2) + "\n")


def get_new_sentence_dict(
    speaker: str,
    start_time: float,
    end_time: float,
    text: str,
    session_id: Optional[str] = None,
    decimal: int = 3,
) -> dict:
    """
    Get a new SegLST style sentence dictionary variable.

    Args:
        speaker (str): The speaker of the sentence.
        start_time (float): The start time of the sentence.
        end_time (float): The end time of the sentence.
        text (str): The text of the sentence.
        session_id (Optional[str]): The session id of the sentence.

    Returns:
        Dict[str, Any]: A new SegLST style sentence dictionary variable.
    """
    # If start_time or end_time is a torch tensor, convert it to a float and round it to 3 decimal places
    if isinstance(start_time, torch.Tensor):
        start_time = start_time.item()
    if isinstance(end_time, torch.Tensor):
        end_time = end_time.item()
    return {
        'speaker': speaker,
        'start_time': round(start_time, decimal),
        'end_time': round(end_time, decimal),
        'words': text.lstrip(),
        'session_id': session_id,
    }


def fix_frame_time_step(cfg: Any, new_tokens: List[str], new_words: List[str], frame_inds_seq: List[int]) -> List[int]:
    """
    Adjust the frame indices sequence to match the length of new tokens.

    This function handles mismatches between the number of tokens and the frame indices sequence.
    It adjusts the frame_inds_seq to ensure it has the same length as new_tokens.

    Args:
        cfg (Any): Configuration object containing logging settings.
        new_tokens (List[str]): List of new tokens.
        new_words (List[str]): List of new words.
        frame_inds_seq (List[int]): List of frame indices.

    Returns:
        List[int]: Adjusted frame indices sequence.
    """
    if len(new_tokens) != len(frame_inds_seq):
        # Sometimes there is a mismatch in the number of tokens between the new tokens and the frame indices sequence.
        if len(frame_inds_seq) > len(new_words):
            # Get unique frame indices sequence
            frame_inds_seq = list(OrderedDict.fromkeys(frame_inds_seq))
            if len(frame_inds_seq) < len(new_tokens):
                deficit = len(new_tokens) - len(frame_inds_seq)
                frame_inds_seq = [frame_inds_seq[0]] * deficit + frame_inds_seq
            elif len(frame_inds_seq) > len(new_tokens):
                deficit = len(frame_inds_seq) - len(new_tokens)
                frame_inds_seq = frame_inds_seq[deficit:]

        elif len(frame_inds_seq) < len(new_tokens):
            deficit = len(new_tokens) - len(frame_inds_seq)
            frame_inds_seq = [frame_inds_seq[0]] * deficit + frame_inds_seq
        if cfg.get("log", True):
            logging.warning(
                f"Length of new token sequence ({len(new_tokens)}) does not match"
                f"the length of frame indices sequence ({len(frame_inds_seq)}). Skipping this chunk."
            )
    return frame_inds_seq


def get_simulated_softmax(cfg, speaker_sigmoid: torch.Tensor) -> torch.Tensor:
    """
    Simulate the softmax operation for speaker diarization.

    Args:
        cfg (Any): Configuration object containing diarization settings.
        speaker_sigmoid (torch.Tensor): Speaker sigmoid values.

    Returns:
        speaker_softmax (torch.Tensor): Speaker softmax values.
    """
    if speaker_sigmoid.ndim != 1:
        raise ValueError(f"Expected 1D tensor for speaker_sigmoid, got shape {speaker_sigmoid.shape}")
    if speaker_sigmoid.shape[0] < cfg.get("max_num_of_spks", 4):
        raise ValueError(
            f"speaker_sigmoid size {speaker_sigmoid.shape[0]} < max_num_of_spks {cfg.get('max_num_of_spks', 4)}"
        )

    speaker_sigmoid = torch.clamp(speaker_sigmoid, min=cfg.get("min_sigmoid_val", 1e-2), max=1)
    sigmoid_sum = speaker_sigmoid.sum()
    if sigmoid_sum == 0:
        logging.warning("speaker_sigmoid sum is zero, returning uniform distribution")
        speaker_softmax = torch.ones_like(speaker_sigmoid) / speaker_sigmoid.shape[0]
    else:
        speaker_softmax = speaker_sigmoid / sigmoid_sum
    speaker_softmax = speaker_softmax.cpu()
    speaker_softmax[cfg.get("max_num_of_spks", 4) :] = 0.0
    return speaker_softmax


def get_word_dict_content_offline(
    cfg: Any,
    word: str,
    word_index: int,
    diar_pred_out: torch.Tensor,
    time_stt_end_tuple: Tuple[int],
    frame_len: float = 0.08,
) -> Dict[str, Any]:
    """
    Generate a dictionary containing word information and speaker diarization results.

    This function processes a single word and its associated tokens to determine
    the start and end frames, speaker, and other relevant information.

    Args:
        cfg (Any): Configuration object containing diarization settings.
        word (str): The word being processed.
        word_index (int): Index of the word in the sequence.
        diar_pred_out (torch.Tensor): Diarization prediction output stream.
        time_stt_end_tuple (int): Local time step offset.

        frame_len (float, optional): Length of each frame in seconds. Defaults to 0.08.

    Returns:
        Dict[str, Any]: A dictionary containing word information and diarization results.
    """
    frame_stt, frame_end = time_stt_end_tuple

    # Edge Cases: Sometimes, repeated token indexs can lead to incorrect frame and speaker assignment.
    if frame_stt == frame_end:
        if frame_stt >= diar_pred_out.shape[0] - 1:
            frame_stt, frame_end = (diar_pred_out.shape[1] - 1, diar_pred_out.shape[0])
        else:
            frame_end = frame_stt + 1

    # Get the speaker based on the frame-wise softmax probabilities.
    stt_p, end_p = max((frame_stt + cfg.get("left_frame_shift", -1)), 0), (frame_end + cfg.get("right_frame_shift", 0))
    speaker_sigmoid = diar_pred_out[stt_p:end_p, :].mean(dim=0)
    speaker_softmax = get_simulated_softmax(cfg, speaker_sigmoid)

    speaker_softmax[cfg.get("max_num_of_spks", 4) :] = 0.0
    spk_id = speaker_softmax.argmax().item()
    stt_sec, end_sec = frame_stt * frame_len, frame_end * frame_len
    word_dict = {
        "word": word,
        "word_index": word_index,
        'frame_stt': frame_stt,
        'frame_end': frame_end,
        'start_time': round(stt_sec, 3),
        'end_time': round(end_sec, 3),
        'speaker': f"speaker_{spk_id}",
        'speaker_softmax': speaker_softmax,
    }
    return word_dict


def get_word_dict_content_online(
    cfg: Any,
    word: str,
    word_index: int,
    diar_pred_out_stream: torch.Tensor,
    token_group: List[str],
    frame_inds_seq: List[int],
    time_step_local_offset: int,
    frame_len: float = 0.08,
) -> Dict[str, Any]:
    """
    Generate a dictionary containing word information and speaker diarization results.

    This function processes a single word and its associated tokens to determine
    the start and end frames, speaker, and other relevant information.

    Args:
        cfg (Any): Configuration object containing diarization settings.
        word (str): The word being processed.
        word_index (int): Index of the word in the sequence.
        diar_pred_out_stream (torch.Tensor): Diarization prediction output stream.
            Dimensions: (num_frames, max_num_of_spks)
        token_group (List[str]): Group of tokens associated with the word.
        frame_inds_seq (List[int]): Sequence of frame indices.
        time_step_local_offset (int): Local time step offset.
        frame_len (float, optional): Length of each frame in seconds. Defaults to 0.08.

    Returns:
        Dict[str, Any]: A dictionary containing word information and diarization results.
    """
    _stt, _end = time_step_local_offset, time_step_local_offset + len(token_group) - 1
    if len(token_group) == 1:
        frame_stt, frame_end = frame_inds_seq[_stt], frame_inds_seq[_stt] + 1
    else:
        try:
            frame_stt, frame_end = frame_inds_seq[_stt], frame_inds_seq[_end]
        except IndexError:
            frame_stt, frame_end = frame_inds_seq[_stt], frame_inds_seq[_stt] + 1

    # Edge Cases: Sometimes, repeated token indexs can lead to incorrect frame and speaker assignment.
    if frame_stt == frame_end:
        if frame_stt >= diar_pred_out_stream.shape[0] - 1:
            frame_stt, frame_end = (diar_pred_out_stream.shape[0] - 1, diar_pred_out_stream.shape[0])
        else:
            frame_end = frame_stt + 1

    # Get the speaker based on the frame-wise softmax probabilities.
    stt_p, end_p = max((frame_stt + cfg.get("left_frame_shift", -1)), 0), (frame_end + cfg.get("right_frame_shift", 0))
    speaker_sigmoid = diar_pred_out_stream[stt_p:end_p, :].mean(dim=0)
    speaker_softmax = get_simulated_softmax(cfg, speaker_sigmoid)

    speaker_softmax[cfg.get("max_num_of_spks", 4) :] = 0.0
    spk_id = speaker_softmax.argmax().item()
    stt_sec, end_sec = frame_stt * frame_len, frame_end * frame_len
    word_dict = {
        "word": word,
        "word_index": word_index,
        'frame_stt': frame_stt,
        'frame_end': frame_end,
        'start_time': round(stt_sec, 3),
        'end_time': round(end_sec, 3),
        'speaker': f"speaker_{spk_id}",
        'speaker_softmax': speaker_softmax,
    }
    return word_dict


def get_multitoken_words(
    cfg, word_and_ts_seq: Dict[str, List], word_seq: List[str], new_words: List[str], fix_prev_words_count: int = 5
) -> Dict[str, List]:
    """
    Fix multi-token words that were not fully captured by the previous chunk window.

    This function compares the words in the current sequence with the previously processed words,
    and updates any multi-token words that may have been truncated in earlier processing.

    Args:
        cfg (DiarizationConfig): Configuration object containing verbose setting.
        word_and_ts_seq (Dict[str, List]): Dictionary containing word sequences and timestamps.
        word_seq (List[str]): List of all words processed so far.
        new_words (List[str]): List of new words in the current chunk.
        fix_prev_words_count (int, optional): Number of previous words to check. Defaults to 5.

    Returns:
        Dict[str, List]: Updated word_and_ts_seq with fixed multi-token words.
    """
    prev_start = max(0, len(word_seq) - fix_prev_words_count - len(new_words))
    prev_end = max(0, len(word_seq) - len(new_words))
    for ct, prev_word in enumerate(word_seq[prev_start:prev_end]):
        if len(word_and_ts_seq["words"]) > fix_prev_words_count - ct:
            saved_word = word_and_ts_seq["words"][-fix_prev_words_count + ct]["word"]
            if len(prev_word) > len(saved_word):
                if cfg.verbose:
                    logging.info(f"[Replacing Multi-token Word]: {saved_word} with {prev_word}")
                word_and_ts_seq["words"][-fix_prev_words_count + ct]["word"] = prev_word
    return word_and_ts_seq


def append_word_and_ts_seq(
    cfg: Any, word_idx_offset: int, word_and_ts_seq: Dict[str, Any], word_dict: Dict[str, Any]
) -> tuple[int, Dict[str, Any]]:
    """
    Append the word dictionary to the word and time-stamp sequence.

    This function updates the word_and_ts_seq dictionary by appending new word information
    and managing the buffered words and speaker count.

    Args:
        cfg (Any): Configuration object containing parameters like word_window.
        word_idx_offset (int): The current word index offset.
        word_and_ts_seq (Dict[str, Any]): Dictionary containing word sequences and related information.
        word_dict (Dict[str, Any]): Dictionary containing information about the current word.

    Returns:
        tuple[int, Dict[str, Any]]: A tuple containing the updated word_idx_offset and word_and_ts_seq.
    """
    word_and_ts_seq["words"].append(word_dict)
    word_and_ts_seq["buffered_words"].append(word_dict)
    word_and_ts_seq["speaker_count_buffer"].append(word_dict["speaker"])
    word_and_ts_seq["word_window_seq"].append(word_dict['word'])

    if len(word_and_ts_seq["words"]) >= cfg.word_window + 1:
        word_and_ts_seq["buffered_words"].pop(0)
        word_and_ts_seq["word_window_seq"].pop(0)
        word_idx_offset = 0

    word_and_ts_seq["speaker_count"] = len(set(word_and_ts_seq["speaker_count_buffer"]))
    return word_idx_offset, word_and_ts_seq


class SpeakerTaggedASR:
    def __init__(
        self,
        cfg,
        asr_model,
        diar_model,
    ):
        # Required configs, models and datasets for inference
        self.cfg = cfg
        if not self.cfg.get("deploy_mode", False):
            if self.cfg.manifest_file:
                self.test_manifest_dict = get_audio_rttm_map(self.cfg.manifest_file)
            elif self.cfg.audio_file is not None:
                uniq_id = get_uniqname_from_filepath(filepath=self.cfg.audio_file)
                self.test_manifest_dict = {
                    uniq_id: {'audio_filepath': self.cfg.audio_file, 'seglst_filepath': None, 'rttm_filepath': None}
                }
            else:
                raise ValueError("One of the audio_file and manifest_file should be non-empty!")
        else:
            self.test_manifest_dict = {
                "streaming_session": {
                    'audio_filepath': 'streaming_session.wav',
                    'seglst_filepath': None,
                    'rttm_filepath': None,
                }
            }
        self.transcribed_speaker_texts = [None] * len(self.test_manifest_dict)

        self.asr_model = asr_model
        self.diar_model = diar_model

        # ASR speaker tagging configs
        self._fix_prev_words_count = cfg.fix_prev_words_count
        self._sentence_render_length = int(self._fix_prev_words_count + cfg.update_prev_words_sentence)
        self._frame_len_sec = 0.08
        self._initial_steps = cfg.ignored_initial_frame_steps
        self._word_and_ts_seq = {}
        self._stt_words = []
        self._frame_hop_length = self.asr_model.encoder.streaming_cfg.valid_out_len
        self._init_transcript_sessions()

        # Multi-instance configs
        self._max_num_of_spks = cfg.get("max_num_of_spks", 4)
        self._offset_chunk_start_time = 0.0
        self._sent_break_sec = cfg.get("sent_break_sec", 5.0)

        self._att_context_size = cfg.att_context_size
        self._nframes_per_chunk = self._att_context_size[1] + 1
        self._cache_gating = cfg.get("cache_gating", False)
        self._cache_gating_buffer_size = cfg.get("cache_gating_buffer_size", 2)
        self._binary_diar_preds = cfg.binary_diar_preds

        self._masked_asr = cfg.get("masked_asr", True)
        self._use_mask_preencode = cfg.get("mask_preencode", False)
        self._single_speaker_mode = cfg.get("single_speaker_mode", False)

        self.instance_manager = MultiTalkerInstanceManager(
            asr_model=self.asr_model,
            diar_model=self.diar_model,
            max_num_of_spks=self.diar_model._cfg.max_num_of_spks,
            batch_size=cfg.batch_size,
            sent_break_sec=self._sent_break_sec,
        )
        self.n_active_speakers_per_stream = self.cfg.max_num_of_spks

    def _init_transcript_sessions(self):
        """
        Initialize the word and time-stamp sequence for each session.
        """
        for uniq_id in self.test_manifest_dict.keys():
            uniq_id = uniq_id.split(".")[0]  # Make sure there is no "." in the uniq_id
            self._word_and_ts_seq[uniq_id] = {
                "words": [],
                "buffered_words": [],
                "token_frame_index": [],
                "offset_count": 0,
                "status": "success",
                "sentences": None,
                "last_word_index": 0,
                "speaker_count": None,
                "transcription": None,
                "max_spk_probs": [],
                "word_window_seq": [],
                "speaker_count_buffer": [],
                "sentence_memory": {},
            }

    def _get_offset_sentence(self, session_trans_dict: Dict[str, Any], offset: int) -> Dict[str, Any]:
        """
        For the very first word in a session, get the offset sentence.

        Args:
            session_trans_dict (dict): Dictionary containing session-related information.
            offset (int): Index of the word for which the offset sentence is needed.

        Returns:
            (Dict): Dictionary containing offset sentence information.
        """
        word_dict = session_trans_dict['words'][offset]
        return {
            'session_id': session_trans_dict['uniq_id'],
            'speaker': word_dict['speaker'],
            'start_time': word_dict['start_time'],
            'end_time': word_dict['end_time'],
            'words': f"{word_dict['word']} ",
        }

    def _get_sentence(self, word_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the sentence for a given word.

        Args:
            word_dict (Dict[str, Any]): Dictionary containing word-related information.
        """
        return {
            'speaker': word_dict['speaker'],
            'start_time': word_dict['start_time'],
            'end_time': word_dict['end_time'],
            'words': '',
        }

    def get_sentences_values(self, session_trans_dict: dict, sentence_render_length: int):
        """
        Get sentences (speaker-turn-level text) for a given session and sentence render length.

        Args:
            session_trans_dict (Dict[str, Any]): Dictionary containing session-related information.
            sentence_render_length (int): Length of the sentences to be generated.

        Returns:
            sentences (List[Dict[str, Any]]): List of sentences in the session.
        """
        stt_word_index = max(0, session_trans_dict['last_word_index'] - sentence_render_length)
        if session_trans_dict['sentences'] is None:
            sentence = self._get_offset_sentence(session_trans_dict=session_trans_dict, offset=0)
            sentences = []
            session_trans_dict['last_word_index'] = stt_word_index
            session_trans_dict['sentence_memory'].update(
                {stt_word_index: (deepcopy(sentences), deepcopy(sentence), sentence['speaker'])}
            )
            prev_speaker = session_trans_dict['words'][stt_word_index]['speaker']
        else:
            (_sentences, _sentence, prev_speaker) = session_trans_dict['sentence_memory'][stt_word_index]
            sentences, sentence = deepcopy(_sentences), deepcopy(_sentence)

        for word_idx in range(stt_word_index + 1, len(session_trans_dict['words'])):
            word_dict = session_trans_dict['words'][word_idx]
            word, end_point = word_dict['word'], word_dict['end_time']
            if word_dict['speaker'] != prev_speaker:
                sentence['words'] = sentence['words'].strip()
                sentences.append(sentence)
                sentence = self._get_sentence(word_dict=session_trans_dict['words'][word_idx])
            else:
                sentence['end_time'] = end_point
            sentence['words'] += word.strip() + ' '
            sentence['words'] = sentence['words']
            sentence['session_id'] = session_trans_dict['uniq_id']
            session_trans_dict['last_word_index'] = word_idx
            prev_speaker = word_dict['speaker']
            session_trans_dict['sentence_memory'][word_idx] = (deepcopy(sentences), deepcopy(sentence), prev_speaker)
        sentence['words'] = sentence['words'].strip()
        sentences.append(sentence)
        session_trans_dict['sentences'] = sentences
        return session_trans_dict

    def merge_transcript_and_speakers(
        self, test_manifest_dict: dict, asr_hypotheses: List[Hypothesis], diar_pred_out: torch.Tensor
    ) -> Tuple[List[str], Dict[str, Dict[str, Any]]]:
        """
        Merge the transcript and speakers and generate real-time scripts if the config is set.

        Args:
            test_manifest_dict (Dict): Dictionary containing test manifest data.
            asr_hypotheses (List[Hypothesis]): List of ASR hypotheses.
            diar_pred_out (torch.Tensor): Diarization prediction output stream.

        Returns:
            transcribed_speaker_texts (List[str]): List of transcribed speaker texts.
            self._word_and_ts_seq (Dict[str, Dict[str, Any]]): Dictionary of word-level dictionaries with uniq_id as key.
        """

        for idx, (uniq_id, _) in enumerate(test_manifest_dict.items()):
            uniq_id = uniq_id.split(".")[0]  # Make sure there is no "." in the uniq_id
            if not len(asr_hypotheses[idx].text) == 0:
                # Get the word-level dictionaries for each word in the chunk
                self._word_and_ts_seq[uniq_id] = self.get_frame_and_words_offline(
                    uniq_id=uniq_id,
                    diar_pred_out=diar_pred_out[idx].squeeze(0),
                    asr_hypothesis=asr_hypotheses[idx],
                    word_and_ts_seq=self._word_and_ts_seq[uniq_id],
                )
                if len(self._word_and_ts_seq[uniq_id]["words"]) > 0:
                    self._word_and_ts_seq[uniq_id] = self.get_sentences_values(
                        session_trans_dict=self._word_and_ts_seq[uniq_id],
                        sentence_render_length=self._sentence_render_length,
                    )
                    if self.cfg.generate_realtime_scripts:
                        self.transcribed_speaker_texts[idx] = print_sentences(
                            sentences=self._word_and_ts_seq[uniq_id]["sentences"],
                            color_palette=get_color_palette(),
                            params=self.cfg,
                        )
                        if not self.cfg.get("deploy_mode", False):
                            write_txt(
                                f'{self.cfg.print_path}'.replace(".sh", f"_{idx}.sh"),
                                self.transcribed_speaker_texts[idx].strip(),
                            )
        return self.transcribed_speaker_texts, self._word_and_ts_seq

    def get_frame_and_words_offline(
        self,
        uniq_id: str,
        diar_pred_out: torch.Tensor,
        asr_hypothesis: Hypothesis,
        word_and_ts_seq: Dict[str, Any],
    ):
        """
        Get the frame and words for each word in the chunk.

        Args:
            uniq_id (str): The unique id of the chunk.
            diar_pred_out (torch.Tensor): Diarization prediction output stream.
            asr_hypothesis (Hypothesis): ASR hypothesis.
            word_and_ts_seq (Dict[str, Any]): Pre-existing word-level dictionaries.

        Returns:
            word_and_ts_seq (Dict[str, Any]): The updated word-level dictionaries with new words.
        """
        word_and_ts_seq['uniq_id'] = uniq_id

        for word_index, hyp_word_dict in enumerate(asr_hypothesis.timestamp['word']):
            time_stt_end_tuple = (hyp_word_dict['start_offset'], hyp_word_dict['end_offset'])
            word_dict = get_word_dict_content_offline(
                cfg=self.cfg,
                word=hyp_word_dict['word'],
                word_index=word_index,
                diar_pred_out=diar_pred_out,
                time_stt_end_tuple=time_stt_end_tuple,
                frame_len=self._frame_len_sec,
            )
            word_and_ts_seq["words"].append(word_dict)
            word_and_ts_seq["speaker_count_buffer"].append(word_dict["speaker"])
            word_and_ts_seq["word_window_seq"].append(word_dict['word'])

        word_and_ts_seq["buffered_words"] = word_and_ts_seq["words"]
        word_and_ts_seq["speaker_count"] = len(set(word_and_ts_seq["speaker_count_buffer"]))
        return word_and_ts_seq

    def get_frame_and_words_online(
        self,
        uniq_id: str,
        step_num: int,
        diar_pred_out_stream: torch.Tensor,
        previous_hypothesis: Hypothesis,
        word_and_ts_seq: Dict[str, Any],
    ):
        """
        Get the frame and words for each word object in the chunk during streaming inference.

        Args:
            uniq_id (str): The unique id of the chunk.
            step_num (int): The step number of the chunk.
            diar_pred_out_stream (torch.Tensor): The diarization prediction output stream.
            previous_hypothesis (Hypothesis): The previous hypothesis.
            word_and_ts_seq (Dict[str, Any]): The word and timestamp sequence.

        Returns:
            word_and_ts_seq (Dict[str, Any]): The word and timestamp sequence.
        """
        offset = step_num * self._frame_hop_length
        word_seq = previous_hypothesis.text.split()
        new_words = word_seq[word_and_ts_seq["offset_count"] :]
        new_token_group = self.asr_model.tokenizer.text_to_tokens(new_words)
        new_tokens = list(itertools.chain(*new_token_group))
        frame_inds_seq = (torch.tensor(previous_hypothesis.timestamp) + offset).tolist()
        frame_inds_seq = fix_frame_time_step(self.cfg, new_tokens, new_words, frame_inds_seq)
        word_and_ts_seq['uniq_id'] = uniq_id

        min_len = min(len(new_words), len(frame_inds_seq))
        for idx in range(min_len):
            word_and_ts_seq["token_frame_index"].append((new_tokens[idx], frame_inds_seq[idx]))
            word_and_ts_seq["offset_count"] += 1

        time_step_local_offset, word_idx_offset = 0, 0
        word_and_ts_seq = get_multitoken_words(
            cfg=self.cfg,
            word_and_ts_seq=word_and_ts_seq,
            word_seq=word_seq,
            new_words=new_words,
            fix_prev_words_count=self._fix_prev_words_count,
        )

        # Get the FIFO queue preds to word_and_ts_seq
        for local_idx, (token_group, word) in enumerate(zip(new_token_group, new_words)):
            word_dict = get_word_dict_content_online(
                cfg=self.cfg,
                word=word,
                word_index=(len(word_and_ts_seq["words"]) + local_idx),
                diar_pred_out_stream=diar_pred_out_stream,
                token_group=token_group,
                frame_inds_seq=frame_inds_seq,
                time_step_local_offset=time_step_local_offset,
                frame_len=self._frame_len_sec,
            )
            # Count the number of speakers in the word window
            time_step_local_offset += len(token_group)
            word_idx_offset, word_and_ts_seq = append_word_and_ts_seq(
                cfg=self.cfg, word_idx_offset=word_idx_offset, word_and_ts_seq=word_and_ts_seq, word_dict=word_dict
            )
        return word_and_ts_seq

    def _add_speaker_transcriptions(
        self,
        transcriptions: list,
        speaker_transcriptions: List[str],
        word_and_ts_seq: Dict[str, Dict[str, Any]],
        test_manifest_dict: dict,
    ) -> Tuple[List[Hypothesis], List[Hypothesis]]:
        """
        Add speaker tagging into the transcriptions generated from an ASR model.

        Args:
            transcriptions (Tuple[List[Hypothesis], List[Hypothesis]]):
                Tuple containing the transcriptions and n-best transcriptions.
            speaker_transcriptions (List[str]):
                List of speaker transcriptions.
            word_and_ts_seq (Dict[str, Dict[str, Any]]):
                Dictionary of word-level dictionaries with uniq_id as key.
            test_manifest_dict (dict):
                Dictionary containing test manifest data.

        Returns:
            Tuple[List[Hypothesis], List[Hypothesis]]: Tuple containing the updated transcriptions with speaker tags.
        """
        trans_hyp, _ = transcriptions
        for sess_idx, (uniq_id, _) in enumerate(test_manifest_dict.items()):
            uniq_id = uniq_id.split(".")[0]  # Make sure there is no "." in the uniq_id
            if speaker_transcriptions[sess_idx] is not None:
                trans_hyp[sess_idx].text = speaker_transcriptions[sess_idx]
            speaker_added_word_dicts = []
            for word_idx, trans_wdict in enumerate(trans_hyp[0].timestamp['word']):
                trans_wdict_copy = deepcopy(trans_wdict)
                trans_wdict_copy['speaker'] = word_and_ts_seq[uniq_id]['words'][word_idx]['speaker']
                speaker_added_word_dicts.append(trans_wdict_copy)
            trans_hyp[sess_idx].timestamp['word'] = speaker_added_word_dicts
            w_count, segment_list = 0, []
            for word_idx, trans_segdict in enumerate(trans_hyp[0].timestamp['segment']):
                words = trans_segdict['segment'].split()
                spk_vote_pool = []
                for word in words:
                    if word.lower() != word_and_ts_seq[uniq_id]['words'][w_count]['word'].lower():
                        raise ValueError(
                            f"Word mismatch: '{word.lower()}' != '{word_and_ts_seq[uniq_id]['words'][w_count]['word'].lower()}' "
                            f"at session {sess_idx}, word count {w_count}."
                        )
                    spk_int = int(word_and_ts_seq[uniq_id]['words'][w_count]['speaker'].split('_')[-1])
                    spk_vote_pool.append(spk_int)
                    w_count += 1
                trans_segdict['speaker'] = f"speaker_{torch.mode(torch.tensor(spk_vote_pool), dim=0).values.item()}"
                segment_list.append(trans_segdict)
            trans_hyp[sess_idx].timestamp['segment'] = segment_list
        transcriptions = (trans_hyp, trans_hyp)
        return transcriptions

    def perform_offline_stt_spk(self, override_cfg: Dict[str, Any]):
        """
        Perform offline STT and speaker diarization on the provided manifest file.

        Args:
            override_cfg (dict): Override configuration parameters.

        Returns:
            transcriptions (Tuple): Tuple containing the speaker-tagged transcripts.
        """
        transcriptions = self.asr_model.transcribe(
            audio=self.cfg.dataset_manifest,
            override_config=override_cfg,
        )
        best_hyp, _ = transcriptions
        _, pred_tensors = self.diar_model.diarize(audio=self.cfg.manifest_file, include_tensor_outputs=True)
        speaker_transcriptions, word_and_ts_seq = self.merge_transcript_and_speakers(
            test_manifest_dict=self.diar_model._diarize_audio_rttm_map,
            asr_hypotheses=best_hyp,
            diar_pred_out=pred_tensors,
        )
        transcriptions = self._add_speaker_transcriptions(
            transcriptions=transcriptions,
            speaker_transcriptions=speaker_transcriptions,
            word_and_ts_seq=word_and_ts_seq,
            test_manifest_dict=self.diar_model._diarize_audio_rttm_map,
        )
        return transcriptions

    def generate_seglst_dicts_from_serial_streaming(self, samples: List[Dict[str, Any]]):
        """
        Generate the seglst dictionary for SegLST format from serial streaming.
        For SegLST format, the session_id is the name of the audio file
        should not contain "." in the name.

        Args:
            samples (List[Dict[str, Any]]): List of samples.
        """
        for sample in samples:
            uniq_id = get_uniqname_from_filepath(sample['audio_filepath']).split('.')[0]
            word_ts_and_seq_dict = self._word_and_ts_seq[uniq_id]
            for sentence_dict in word_ts_and_seq_dict['sentences']:
                session_id = word_ts_and_seq_dict['uniq_id'].split('.')[0]
                seglst_dict = get_new_sentence_dict(
                    speaker=sentence_dict['speaker'],
                    start_time=float(sentence_dict['start_time']),
                    end_time=float(sentence_dict['end_time']),
                    text=sentence_dict["words"],
                    session_id=session_id,
                )
                if seglst_dict['words'].strip() != "":
                    self.instance_manager.seglst_dict_list.append(seglst_dict)
        return self.instance_manager.seglst_dict_list

    def generate_seglst_dicts_from_parallel_streaming(self, samples: List[Dict[str, Any]]):
        """
        Generate the seglst dictionary for SegLST format from parallel streaming.
        For SegLST format, the session_id is the name of the audio file
        should not contain "." in the name.

        Args:
            samples (List[Dict[str, Any]]): List of samples.
        """
        self.instance_manager.previous_asr_states.extend(self.instance_manager.batch_asr_states)
        for sample, asr_state in zip(samples, self.instance_manager.previous_asr_states):
            audio_filepath = sample["audio_filepath"]
            uniq_id = os.path.basename(audio_filepath).split('.')[0]
            seglsts = []
            for seg in asr_state.seglsts:
                a_seg_dict = get_new_sentence_dict(
                    speaker=seg['speaker'],
                    start_time=seg['start_time'],
                    end_time=seg['end_time'],
                    text=seg['words'],
                    session_id=uniq_id,
                )
                if a_seg_dict['words'].strip() != "":
                    seglsts.append(a_seg_dict)
            seglsts = sorted(seglsts, key=lambda x: x['start_time'])
            self.instance_manager.seglst_dict_list.extend(seglsts)
        return self.instance_manager.seglst_dict_list

    def _find_active_speakers(self, diar_preds: torch.Tensor, n_active_speakers_per_stream: int) -> List[List[int]]:
        """
        Find the active speakers from the diar prediction output.

        Args:
            diar_preds (torch.Tensor): The diar prediction output.
            n_active_speakers_per_stream (int): The number of active speakers per stream.

        Returns:
            speaker_ids_list (List[List[int]]): The list of active speakers for each stream.
        """
        if diar_preds.ndim != 3:
            raise ValueError(f"diar_preds must be 3D (B, T, N), got shape {diar_preds.shape}")
        if n_active_speakers_per_stream > diar_preds.shape[2]:
            raise ValueError(
                f"n_active_speakers_per_stream ({n_active_speakers_per_stream}) "
                f"> available speakers ({diar_preds.shape[2]})"
            )
        max_probs = torch.max(diar_preds, dim=1).values  # (B, T, N) --> (B, N)
        top_values, top_indices = torch.topk(max_probs, k=n_active_speakers_per_stream, dim=1)
        masks = top_values > 0.5

        speaker_ids_list = []
        for speaker_ids, mask in zip(top_indices, masks):
            speaker_ids_list.append(sorted(speaker_ids[mask].tolist()))
        return speaker_ids_list

    def forward_pre_encoded(
        self, audio_signal: torch.Tensor, length: torch.Tensor, drop_extra_pre_encoded: int = 0
    ) -> None:
        """
        Forward the pre-encoded features through the ASR model.

        Args:
            audio_signal (torch.Tensor): The audio signal.
            length (torch.Tensor): The length of the audio signal.
            drop_extra_pre_encoded (int): The number of extra pre-encoded tokens to drop.

        Returns:
            audio_signal (torch.Tensor): The pre-encoded audio signal.
            length (torch.Tensor): The length of the pre-encoded audio signal.
        """
        audio_signal = torch.transpose(audio_signal, 1, 2)  # (B, T, D) -> (B, D, T)

        audio_signal, length = self.asr_model.encoder.pre_encode(x=audio_signal, lengths=length)
        length = length.to(torch.int64)
        # `self.streaming_cfg` is set by setup_streaming_cfg(), called in the init
        if drop_extra_pre_encoded:
            audio_signal = audio_signal[:, drop_extra_pre_encoded:, :]
            length = (length - drop_extra_pre_encoded).clamp(min=0)
        return audio_signal, length

    def mask_features(
        self, chunk_audio: torch.Tensor, mask: torch.Tensor, threshold: float = 0.5, mask_value: float = -16.6355
    ) -> torch.Tensor:
        """
        Mask the features of the chunk audio.

        Args:
            chunk_audio (torch.Tensor): The chunk audio.
            mask (torch.Tensor): The mask.
            threshold (float): The threshold for the mask.
            mask_value (float): The value for the masked audio.

        Returns:
            masked_chunk_audio (torch.Tensor): The masked chunk audio.
        """
        if chunk_audio.ndim != 3:
            raise ValueError(
                f"chunk_audio must be 3D (B, C, T), got {chunk_audio.ndim}D with shape {chunk_audio.shape}"
            )
        if mask.ndim != 2:
            raise ValueError(f"mask must be 2D (B, T), got {mask.ndim}D with shape {mask.shape}")
        if chunk_audio.shape[0] != mask.shape[0]:
            raise ValueError(f"Batch size mismatch: chunk_audio={chunk_audio.shape[0]}, mask={mask.shape[0]}")
        mask = (mask > threshold).float()
        mask = mask.unsqueeze(-1).repeat(1, 1, 8).flatten(1, 2)

        if mask.shape[1] > chunk_audio.shape[2]:
            if self.cfg.get("log", False):
                logging.warning(f"Mask shape {mask.shape} is greater than chunk_audio shape {chunk_audio.shape}")
            mask = mask[:, : chunk_audio.shape[2]]
        elif mask.shape[1] < chunk_audio.shape[2]:
            if self.cfg.get("log", False):
                logging.warning(f"Mask shape {mask.shape} is less than chunk_audio shape {chunk_audio.shape}")
            mask = torch.nn.functional.pad(mask, (chunk_audio.shape[2] - mask.shape[1], 0), mode='constant', value=0)

        masked_chunk_audio = chunk_audio * mask.unsqueeze(1)
        masked_chunk_audio[torch.where(chunk_audio == 0)] = mask_value

        return masked_chunk_audio

    def mask_preencode(self, chunk_audio: torch.Tensor, mask: torch.Tensor, threshold: float = 0.5) -> torch.Tensor:
        """
        Mask the pre-encoded features of the chunk audio.

        Args:
            chunk_audio (torch.Tensor): The chunk audio.
            mask (torch.Tensor): The mask.
            threshold (float): The threshold for the mask.

        Returns:
            masked_chunk_audio (torch.Tensor): The masked chunk audio.
        """
        mask = (mask > threshold).float()

        if mask.shape[1] > chunk_audio.shape[1]:
            logging.warning(f"Mask shape {mask.shape} is greater than chunk_audio shape {chunk_audio.shape}")
            mask = mask[:, : chunk_audio.shape[1]]
        elif mask.shape[1] < chunk_audio.shape[1]:
            logging.warning(f"Mask shape {mask.shape} is less than chunk_audio shape {chunk_audio.shape}")
            mask = torch.nn.functional.pad(mask, (chunk_audio.shape[1] - mask.shape[1], 0), mode='constant', value=0)

        masked_chunk_audio = chunk_audio * mask.unsqueeze(-1)

        return masked_chunk_audio

    def get_diar_pred_out_stream(self, step_num):
        """
        Get the diar prediction output stream for the given step number.

        Args:
            step_num (int): the step number

        Returns:
            new_diar_pred_out_stream (torch.Tensor): the diar prediction output stream for the given step number
            new_chunk_preds (torch.Tensor): the diar prediction output stream for the given step number
        """
        start_frame_idx = step_num * self._nframes_per_chunk
        end_frame_idx = start_frame_idx + self._nframes_per_chunk
        new_diar_pred_out_stream = self.diar_model.rttms_mask_mats[:, :end_frame_idx]
        new_chunk_preds = new_diar_pred_out_stream[:, start_frame_idx:end_frame_idx]
        return new_diar_pred_out_stream, new_chunk_preds

    @measure_eta
    def perform_serial_streaming_stt_spk(
        self,
        step_num: int,
        chunk_audio: torch.Tensor,
        chunk_lengths: torch.Tensor,
        is_buffer_empty: bool,
        drop_extra_pre_encoded: int,
    ):
        """
        Perform the serial streaming inference.
        Serial streaming inference deploys a single ASR model instance to transcribe multiple speakers in a chunk.
        All the updates are done to the instance manager in a `SpeakerTaggedASR` class instance.

        Args:
            step_num (int): The step number of the chunk.
            chunk_audio (torch.Tensor): The chunk audio.
            chunk_lengths (torch.Tensor): The length of the chunk audio.
            is_buffer_empty (bool): Whether the buffer is empty.
            drop_extra_pre_encoded (int): The number of extra pre-encoded tokens to drop.
        """
        # Initialize the instance manager with the batch size of the chunk audio.
        if step_num == 0:
            self.instance_manager.reset(batch_size=chunk_audio.shape[0])
            self.instance_manager.to(chunk_audio.device)

        # This part exists for compatibility with the parallel streaming inference.
        self.instance_manager.get_active_speakers_info(
            active_speakers=[[0] for _ in range(chunk_audio.shape[0])],
            chunk_audio=chunk_audio,
            chunk_lengths=chunk_lengths,
        )

        (
            asr_pred_out_stream,
            _,
            cache_last_channel,
            cache_last_time,
            cache_last_channel_len,
            previous_hypotheses,
        ) = self.asr_model.conformer_stream_step(
            processed_signal=chunk_audio,
            processed_signal_length=chunk_lengths,
            cache_last_channel=self.instance_manager.active_cache_last_channel,
            cache_last_time=self.instance_manager.active_cache_last_time,
            cache_last_channel_len=self.instance_manager.active_cache_last_channel_len,
            previous_hypotheses=self.instance_manager.active_previous_hypotheses,
            previous_pred_out=self.instance_manager.active_asr_pred_out_stream,
            keep_all_outputs=is_buffer_empty,
            drop_extra_pre_encoded=drop_extra_pre_encoded,
            return_transcription=True,
        )

        if self.diar_model.rttms_mask_mats is None:

            new_streaming_state, diar_pred_out_stream = self.diar_model.forward_streaming_step(
                processed_signal=chunk_audio.transpose(1, 2),
                processed_signal_length=chunk_lengths,
                streaming_state=self.instance_manager.diar_states.streaming_state,
                total_preds=self.instance_manager.diar_states.diar_pred_out_stream,
                drop_extra_pre_encoded=drop_extra_pre_encoded,
            )
            self.instance_manager.update_diar_state(
                diar_pred_out_stream=diar_pred_out_stream,
                previous_chunk_preds=diar_pred_out_stream[:, -self._nframes_per_chunk :],
                diar_streaming_state=new_streaming_state,
            )
        else:
            _, new_chunk_preds = self.get_diar_pred_out_stream(step_num)
            diar_pred_out_stream = new_chunk_preds

        # Apply max number of speakers
        diar_pred_out_stream[:, :, self._max_num_of_spks :] = 0.0

        for idx, (uniq_id, _) in enumerate(self.test_manifest_dict.items()):
            if not (len(previous_hypotheses[idx].text) == 0 and step_num <= self._initial_steps):
                # Get the word-level dictionaries for each word in the chunk
                self._word_and_ts_seq[uniq_id] = self.get_frame_and_words_online(
                    uniq_id=uniq_id,
                    step_num=step_num,
                    diar_pred_out_stream=diar_pred_out_stream[idx, :, :],
                    previous_hypothesis=previous_hypotheses[idx],
                    word_and_ts_seq=self._word_and_ts_seq[uniq_id],
                )
                if len(self._word_and_ts_seq[uniq_id]["words"]) > 0:
                    self._word_and_ts_seq[uniq_id] = self.get_sentences_values(
                        session_trans_dict=self._word_and_ts_seq[uniq_id],
                        sentence_render_length=self._sentence_render_length,
                    )
                    if self.cfg.get("generate_realtime_scripts", True):
                        self.transcribed_speaker_texts[idx] = print_sentences(
                            sentences=self._word_and_ts_seq[uniq_id]["sentences"],
                            color_palette=get_color_palette(),
                            params=self.cfg,
                        )
                        if not self.cfg.get("deploy_mode", False):
                            write_txt(
                                f'{self.cfg.get("print_path", "./print_script.sh")}'.replace(".sh", f"_{idx}.sh"),
                                self.transcribed_speaker_texts[idx].strip(),
                            )

        for batch_idx in range(chunk_audio.shape[0]):
            self.instance_manager.update_asr_state(
                batch_idx,
                speaker_id=0,
                cache_last_channel=cache_last_channel[:, batch_idx],
                cache_last_time=cache_last_time[:, batch_idx],
                cache_last_channel_len=cache_last_channel_len[batch_idx],
                previous_hypotheses=previous_hypotheses[batch_idx],
                previous_pred_out=asr_pred_out_stream[batch_idx],
            )
        return self.transcribed_speaker_texts

    @measure_eta
    def perform_parallel_streaming_stt_spk(
        self,
        step_num,
        chunk_audio,
        chunk_lengths,
        is_buffer_empty,
        drop_extra_pre_encoded,
    ):
        """
        Perform the parallel streaming inference.
        Parallel streaming inference deploys multiple ASR model instances to transcribe multiple speakers in a chunk.
        All the updates are done to the instance manager in a `SpeakerTaggedASR` class instance.

        Args:
            step_num (int): The step number of the chunk.
            chunk_audio (torch.Tensor): The chunk audio.
            chunk_lengths (torch.Tensor): The length of the chunk audio.
            is_buffer_empty (bool): Whether the buffer is empty.
            drop_extra_pre_encoded (int): The number of extra pre-encoded tokens to drop.
        """
        # Initialize the instance manager with the batch size of the chunk audio.
        if step_num == 0:
            self._offset_chunk_start_time = 0
            self.instance_manager.reset(batch_size=chunk_audio.shape[0])
            self.instance_manager.to(chunk_audio.device)

        # Step 2: diarize or get GT rttms
        if self.diar_model.rttms_mask_mats is None:
            new_streaming_state, new_diar_pred_out_stream = self.diar_model.forward_streaming_step(
                processed_signal=chunk_audio.transpose(1, 2),
                processed_signal_length=chunk_lengths,
                streaming_state=self.instance_manager.diar_states.streaming_state,
                total_preds=self.instance_manager.diar_states.diar_pred_out_stream,
                drop_extra_pre_encoded=drop_extra_pre_encoded,
            )
            new_chunk_preds = new_diar_pred_out_stream[:, -self._nframes_per_chunk :]

        else:
            new_diar_pred_out_stream, new_chunk_preds = self.get_diar_pred_out_stream(step_num)
            new_streaming_state = self.instance_manager.diar_states.streaming_state

        # Apply max number of speakers
        new_diar_pred_out_stream[:, :, self._max_num_of_spks :] = 0.0

        # Step 3: update diar states
        self.instance_manager.update_diar_state(
            diar_pred_out_stream=new_diar_pred_out_stream,
            previous_chunk_preds=new_chunk_preds,
            diar_streaming_state=new_streaming_state,
        )

        # For a session, if no second speaker is detected,
        # the spk_targets will be set to all ones in the single speaker mode
        if self._single_speaker_mode:
            if self._max_num_of_spks == 1:
                is_single_speaker = [True] * chunk_audio.shape[0]
            else:
                is_single_speaker = (new_diar_pred_out_stream > 0.5).any(1).sum(-1) <= 1.0
            for i in range(chunk_audio.shape[0]):
                if is_single_speaker[i]:
                    new_diar_pred_out_stream[i, :, 0] = 1.0
                    new_diar_pred_out_stream[i, :, 1:] = 0.0

        # Step 4: find active speakers
        diar_chunk_preds = new_diar_pred_out_stream[:, -self._nframes_per_chunk * self._cache_gating_buffer_size :]
        if self._cache_gating:
            active_speakers = self._find_active_speakers(
                diar_chunk_preds, n_active_speakers_per_stream=self.n_active_speakers_per_stream
            )
        else:
            active_speakers = [list(range(self.n_active_speakers_per_stream)) for _ in range(chunk_audio.shape[0])]

        if (self._masked_asr and self._use_mask_preencode) or not self._masked_asr:
            chunk_audio, chunk_lengths = self.forward_pre_encoded(chunk_audio, chunk_lengths, drop_extra_pre_encoded)
            bypass_pre_encode = True
        else:
            bypass_pre_encode = False

        # Step 5: generate instance for active speakers
        (
            active_chunk_audio,
            active_chunk_lengths,
            active_speaker_targets,
            inactive_speaker_targets,
        ) = self.instance_manager.get_active_speakers_info(
            active_speakers=active_speakers,
            chunk_audio=chunk_audio,
            chunk_lengths=chunk_lengths,
        )

        # skip current chunk if no active speakers are found
        if active_chunk_audio is None:
            return

        # Step 6:
        # 1) mask the non-active speakers for masked ASR; or
        # 2) set speaker targets for multitalker ASR
        if self._masked_asr:
            if self._use_mask_preencode:
                active_chunk_audio = self.mask_preencode(chunk_audio=active_chunk_audio, mask=active_speaker_targets)
            else:
                active_chunk_audio = self.mask_features(chunk_audio=active_chunk_audio, mask=active_speaker_targets)
        else:
            if self._binary_diar_preds:
                active_speaker_targets = (active_speaker_targets > 0.5).float()
                inactive_speaker_targets = (inactive_speaker_targets > 0.5).float()
            self.asr_model.set_speaker_targets(active_speaker_targets, inactive_speaker_targets)

        # Step 7: ASR forward pass for active speakers
        (
            pred_out_stream,
            _,
            cache_last_channel,
            cache_last_time,
            cache_last_channel_len,
            previous_hypotheses,
        ) = self.asr_model.conformer_stream_step(
            processed_signal=active_chunk_audio,
            processed_signal_length=active_chunk_lengths,
            cache_last_channel=self.instance_manager.active_cache_last_channel,
            cache_last_time=self.instance_manager.active_cache_last_time,
            cache_last_channel_len=self.instance_manager.active_cache_last_channel_len,
            keep_all_outputs=is_buffer_empty,
            previous_hypotheses=self.instance_manager.active_previous_hypotheses,
            previous_pred_out=self.instance_manager.active_asr_pred_out_stream,
            drop_extra_pre_encoded=drop_extra_pre_encoded,
            return_transcription=True,
            bypass_pre_encode=bypass_pre_encode,
        )

        # Step 8: update ASR states
        active_id = 0
        for batch_idx, speaker_ids in enumerate(active_speakers):
            for speaker_id in speaker_ids:
                self.instance_manager.update_asr_state(
                    batch_idx,
                    speaker_id,
                    cache_last_channel[:, active_id],
                    cache_last_time[:, active_id],
                    cache_last_channel_len[active_id],
                    previous_hypotheses[active_id],
                    pred_out_stream[active_id],
                )
                active_id += 1

        # Step 9: update seglsts with timestamps
        self.instance_manager.update_seglsts(offset=self._offset_chunk_start_time)
        self._offset_chunk_start_time += self._nframes_per_chunk * self._frame_len_sec

        if self.cfg.get("generate_realtime_scripts", True):
            for session_idx in self.cfg.get("print_sample_indices", [0]):
                asr_state = self.instance_manager.batch_asr_states[session_idx]
                self.transcribed_speaker_texts[session_idx] = print_sentences(
                    sentences=asr_state.seglsts, color_palette=get_color_palette(), params=self.cfg
                )
                if not self.cfg.get("deploy_mode", False):
                    write_txt(
                        f'{self.cfg.get("print_path", "./print_script.sh").replace(".sh", f"_{session_idx}.sh")}',
                        self.transcribed_speaker_texts[session_idx].strip(),
                    )
        return self.transcribed_speaker_texts


class MultiTalkerInstanceManager:
    """
    For multi-talker inference, we need to manage the information per speaker.
    Each sample in a batch can be considered as a multi-talker instance,
    and each instance may contain multiple speakers, which is the real
    batch size for inference. If there are at most N speakers and the batch
    size is B, then the real batch size for inference is at most B * N.
    """

    class ASRState:
        """
        ASR state for each instance.
        1. In parallel mode, each instance handles each potential speaker.
        2. In serial mode, each instance handles one session.

        The goal of ASR-State class is to handle the ASR cache state between streaming steps.
        The ASR-states required to perform streaming inference are all included in this class.
        """

        def __init__(
            self,
            max_num_of_spks: int = 4,
            frame_len_sec: float = 0.08,
            sent_break_sec: float = 5.0,
            uppercase_first_letter: bool = True,
        ):
            """
            Initialize the ASR-State class with the initial parameters.

            Args:
                max_num_of_spks (int): The maximum number of speakers.
                frame_len_sec (float): The length of the frame in seconds.
                sent_break_sec (float): The minimum time gap between two sentences in seconds.
            """
            # Initialize the ASR state with the initial parameters.
            self.speakers: Optional[List[str]] = None
            self.cache_last_channel = None
            self.cache_last_time = None
            self.cache_last_channel_len = None
            self.previous_hypothesis = None
            self.previous_pred_out = None

            self.max_num_of_spks = max_num_of_spks

            self._frame_len_sec = frame_len_sec
            self._sent_break_sec = sent_break_sec
            self._uppercase_first_letter = uppercase_first_letter
            self._speaker_wise_sentences = {}
            self._prev_history_speaker_texts = ["" for _ in range(self.max_num_of_spks)]

            self.seglsts = []

        def _reset_speaker_wise_sentences(self):
            """
            Reset the speaker-wise sentences which will be used to generate the SegLST transcription outputs.
            """
            self._speaker_wise_sentences = {}
            self._prev_history_speaker_texts = ["" for _ in range(self.max_num_of_spks)]

        def reset(self, asr_cache_state: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
            """
            Reset the ASR state.

            Args:
                asr_cache_state (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The ASR cache state.
                    - cache_last_channel (torch.Tensor): The cache last channel.
                    - cache_last_time (torch.Tensor): The cache last time.
                    - cache_last_channel_len (torch.Tensor): The cache last channel length.
            """
            self.speakers = [0]
            self.cache_last_channel, self.cache_last_time, self.cache_last_channel_len = asr_cache_state
            self.previous_hypothesis = [None]
            self.previous_pred_out = [None]
            self.seglsts = []
            self._speaker_wise_sentences = {}
            self._prev_history_speaker_texts = ["" for _ in range(self.max_num_of_spks)]

        def update_asr_state(
            self,
            speaker_id,
            cache_last_channel,
            cache_last_time,
            cache_last_channel_len,
            previous_hypothesis,
            previous_pred_out,
        ):
            """
            Update the ASR state with the new ASR cache state.
            This function should be called at every streaming step to update the ASR cache state.

            Args:
                speaker_id (int): The speaker id.
                cache_last_channel (torch.Tensor): The cache last channel.
                cache_last_time (torch.Tensor): The cache last time.
                cache_last_channel_len (torch.Tensor): The cache last channel length.
                previous_hypothesis (Hypothesis): The previous hypothesis.
                previous_pred_out (torch.Tensor): The previous prediction output.
            """
            self.cache_last_channel[:, speaker_id] = cache_last_channel
            self.cache_last_time[:, speaker_id] = cache_last_time
            self.cache_last_channel_len[speaker_id] = cache_last_channel_len
            self.previous_hypothesis[speaker_id] = previous_hypothesis
            self.previous_pred_out[speaker_id] = previous_pred_out

        def to(self, device):
            """
            Override the to method to move the ASR state to the device.

            Args:
                device (torch.device): The device to move the ASR state to.
            """
            self.cache_last_channel = self.cache_last_channel.to(device)
            self.cache_last_time = self.cache_last_time.to(device)
            self.cache_last_channel_len = self.cache_last_channel_len.to(device)

        def get_speakers(self):
            """
            Get the speaker ids (int) for each instance.
            This function is used for serial streaming mode.
            """
            return self.speakers

        def add_speaker(self, speaker_id: int, asr_cache_state: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
            """
            Add a speaker index and its initial cache state to the ASR state.

            Args:
                speaker_id (int): The speaker id.
                asr_cache_state (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The ASR cache state.
            """
            self.speakers.append(speaker_id)
            cache_last_channel, cache_last_time, cache_last_channel_len = asr_cache_state
            self.cache_last_channel = torch.cat([self.cache_last_channel, cache_last_channel], dim=1)
            self.cache_last_time = torch.cat([self.cache_last_time, cache_last_time], dim=1)
            self.cache_last_channel_len = torch.cat([self.cache_last_channel_len, cache_last_channel_len], dim=0)
            self.previous_hypothesis.append(None)
            self.previous_pred_out.append(None)

        def _update_last_sentence(self, spk_idx: int, end_time: float, diff_text: str):
            """
            Update the end time of the last sentence for a speaker.

            Args:
                spk_idx (int): The speaker id.
                end_time (float): The end time of the last sentence.
                diff_text (str): The difference text.
            """
            if end_time is not None:
                self._speaker_wise_sentences[spk_idx][-1]['end_time'] = end_time
            new_words = self._speaker_wise_sentences[spk_idx][-1]['words'] + diff_text
            self._speaker_wise_sentences[spk_idx][-1]['words'] = new_words.strip()

        def _is_new_text(self, spk_idx: int, text: str):
            """
            Check if the text is new for a speaker.

            Args:
                spk_idx (int): The speaker id.
                text (str): The text.
            """
            if text is None or text == self._prev_history_speaker_texts[spk_idx]:
                return None
            else:
                # Get the difference between the current text and the previous text
                if self._prev_history_speaker_texts[spk_idx] in text:
                    return text.replace(self._prev_history_speaker_texts[spk_idx], "")
                else:
                    return text.strip()

        def _compute_hypothesis_timestamps(self, hypothesis: Hypothesis, offset: float) -> Tuple[float, float, bool]:
            """
            Compute start and end timestamps for a hypothesis based on available timing information.

            This method calculates the temporal boundaries of a speech hypothesis, prioritizing
            frame-level timestamps when available. When timestamps are not available, it falls
            back to computing timing based on the hypothesis length.

            Args:
                hypothesis (Hypothesis): The ASR hypothesis object containing either frame-level
                offset (float): The time offset (in seconds) to add to the computed timestamps,
                    typically representing the start time of the current audio chunk.

            Returns:
                Tuple[float, float, bool]: A tuple containing:
                    - start_time (float): The absolute start time of the hypothesis in seconds
                    - end_time (float): The absolute end time of the hypothesis in seconds
                    - sep_flag (bool): A flag indicating whether timing was computed from length
                        rather than timestamps.

            Note:
                The end_time calculation from timestamps adds 1 to the last timestamp to account
                for the full duration of the final frame.
            """
            sep_flag = False
            if len(hypothesis.timestamp) > 0:
                start_time = offset + (hypothesis.timestamp[0]) * self._frame_len_sec
                end_time = offset + (hypothesis.timestamp[-1] + 1) * self._frame_len_sec
            else:
                start_time = offset
                end_time = offset + hypothesis.length.item() * self._frame_len_sec
                sep_flag = True

            return start_time, end_time, sep_flag

        def update_sessionwise_seglsts_for_parallel(self, offset: float):
            """
            Update the seglsts for the parallel mode streaming.
            Note that this function is NOT used for serial mode streaming.

            Args:
                offset (float): The offset in seconds.
                                This is usally the start time of the current audio chunk.
            """
            valid_speakers = set()
            for spk_idx in self.get_speakers():
                hypothesis = self.previous_hypothesis[spk_idx]
                if hypothesis is None:
                    continue
                valid_speakers.add(spk_idx)

                if spk_idx not in self._speaker_wise_sentences:
                    self._speaker_wise_sentences[spk_idx] = []

                diff_text = self._is_new_text(spk_idx=spk_idx, text=hypothesis.text)
                if diff_text is not None:

                    start_time, end_time, sep_flag = self._compute_hypothesis_timestamps(
                        hypothesis=hypothesis, offset=offset
                    )

                    # Get the last end time of the previous sentence or None if no sentences are present
                    if len(self._speaker_wise_sentences[spk_idx]) > 0:
                        last_end_time = self._speaker_wise_sentences[spk_idx][-1]['end_time']
                    else:
                        last_end_time = 0.0

                    # Case 1 - If start_tiime is greater than end_time + sent_break_sec, then we need to add the sentence
                    if sep_flag or (last_end_time == 0.0 or start_time > last_end_time + self._sent_break_sec):
                        stripped_text = diff_text.strip()
                        if len(stripped_text) > 0 and stripped_text[0] in ['.', ',', '?', '!']:
                            # This handles the case where the first character should be assigned to the previous sentence.
                            the_first_char, diff_text = stripped_text[0], stripped_text[1:]
                            self._update_last_sentence(spk_idx=spk_idx, end_time=None, diff_text=the_first_char)

                        # Add the sentence to the speaker-wise sentences only if the text is not empty.
                        a_seg_dict = get_new_sentence_dict(
                            speaker=f"speaker_{spk_idx}", start_time=start_time, end_time=end_time, text=diff_text
                        )
                        if a_seg_dict['words'].strip() != "":
                            if self._uppercase_first_letter:
                                split_words = a_seg_dict['words'].split()
                                split_words[0] = split_words[0].capitalize()
                                a_seg_dict['words'] = ' '.join(split_words)
                            self._speaker_wise_sentences[spk_idx].append(a_seg_dict)
                    # Case 2 - If start_time is less than end_time + sent_break_sec, then we need to update the end_time
                    else:
                        self._update_last_sentence(spk_idx=spk_idx, end_time=end_time, diff_text=diff_text)

                # Update the previous history of the speaker text
                if hypothesis.text is not None:
                    self._prev_history_speaker_texts[spk_idx] = hypothesis.text

            self.seglsts = []

            # Merge all sentences for each speaker but sort by start_time
            for spk_idx in valid_speakers:
                self.seglsts.extend(self._speaker_wise_sentences[spk_idx])

            # Finally, sort the seglsts by start_time
            self.seglsts = sorted(self.seglsts, key=lambda x: x['start_time'])

    class DiarState:
        """
        Diar state for each diarization instance.
        There is no difference between serial and parallel mode for the diarization state.
        The goal of Diar-State class is to handle the diarization cache state between streaming steps.
        """

        def __init__(self, batch_size: int = 1, max_num_of_spks: int = 4):
            """
            Initialize the Diar-State class with the initial parameters.

            Args:
                batch_size (int): The batch size.
                max_num_of_spks (int): The maximum number of speakers.
            """
            self.batch_size = batch_size
            self.max_num_of_spks = max_num_of_spks
            self.diar_pred_out_stream = None
            self.previous_chunk_preds = None
            self.streaming_state = None

        def reset(self, diar_streaming_state: StreamingSortformerState):
            self.diar_pred_out_stream = torch.zeros((self.batch_size, 0, self.max_num_of_spks))
            self.previous_chunk_preds = torch.zeros((self.batch_size, 0, self.max_num_of_spks))
            self.streaming_state = diar_streaming_state

        def to(self, device):
            self.diar_pred_out_stream = self.diar_pred_out_stream.to(device)
            self.previous_chunk_preds = self.previous_chunk_preds.to(device)
            self.streaming_state.to(device)

    def __init__(
        self,
        asr_model=None,
        diar_model=None,
        batch_size: int = 1,
        max_num_of_spks: int = 4,
        sent_break_sec: float = 5.0,
    ):
        """
        Initialize the MultiTalkerInstanceManager class with the initial parameters.

        Args:
            asr_model: The ASR model.
            diar_model: The diarization model.
            batch_size (int): The batch size for ASR.
                1. For parallel mode, this is the number of potential speakers
                multiplied by the session counts.
                2. For serial mode, this is the number of sessions.
            max_num_of_spks (int): The maximum number of speakers.
        """
        self.asr_model = asr_model
        self.diar_model = diar_model

        self.batch_size = batch_size
        self.max_num_of_spks = max_num_of_spks
        self._sent_break_sec = sent_break_sec

        # ASR state bank
        self.batch_asr_states = []
        self.previous_asr_states = []

        # Diar states
        self.diar_states = None

        # SegLST output list
        self.seglst_dict_list = []

        # Active speaker buffer lists
        self._active_chunk_audio: List[torch.Tensor] = []
        self._active_chunk_lengths: List[torch.Tensor] = []
        self._active_speaker_targets: List[torch.Tensor] = []
        self._inactive_speaker_targets: List[torch.Tensor] = []
        self._active_previous_hypotheses: List[Hypothesis] = []
        self._active_asr_pred_out_stream: List[torch.Tensor] = []
        self._active_cache_last_channel: List[torch.Tensor] = []
        self._active_cache_last_time: List[torch.Tensor] = []
        self._active_cache_last_channel_len: List[torch.Tensor] = []

        # Active speaker attributes
        self.active_previous_hypotheses: Optional[List[Hypothesis]] = None
        self.active_asr_pred_out_stream: Optional[List[torch.Tensor]] = None
        self.active_cache_last_channel: Optional[torch.Tensor] = None
        self.active_cache_last_time: Optional[torch.Tensor] = None
        self.active_cache_last_channel_len: Optional[torch.Tensor] = None

    def _reset_active_speaker_buffers(self):
        """
        Reset the active speaker buffers need to update the active speaker information.
        """
        self._active_chunk_audio = []
        self._active_chunk_lengths = []
        self._active_speaker_targets = []
        self._inactive_speaker_targets = []
        self._active_previous_hypotheses = []

        self._active_asr_pred_out_stream = []
        self._active_cache_last_channel = []
        self._active_cache_last_time = []
        self._active_cache_last_channel_len = []

    def reset(self, batch_size: Optional[int] = None, max_num_of_spks: Optional[int] = None):
        """
        Reset the active speaker buffers need to update the active speaker information.

        Args:
            batch_size (Optional[int]): The batch size.
            max_num_of_spks (Optional[int]): The maximum number of speakers.
        """
        if batch_size is not None:
            self.batch_size = batch_size
        if max_num_of_spks is not None:
            self.max_num_of_spks = max_num_of_spks

        if len(self.batch_asr_states) > 0:
            self.previous_asr_states.extend(deepcopy(self.batch_asr_states))
        self.batch_asr_states = [
            self.ASRState(self.max_num_of_spks, sent_break_sec=self._sent_break_sec) for _ in range(self.batch_size)
        ]

        for i in range(self.batch_size):
            self.batch_asr_states[i].reset(self.asr_model.encoder.get_initial_cache_state(batch_size=1))

        self.diar_states = self.DiarState(batch_size=self.batch_size, max_num_of_spks=self.max_num_of_spks)
        self.diar_states.reset(self.diar_model.sortformer_modules.init_streaming_state(batch_size=self.batch_size))

        self.seglst_dict_list = []

    def add_speaker(self, batch_idx: int, speaker_id: int):
        """
        Add a speaker index and its initial cache state to the ASR state.

        Args:
            batch_idx (int): The batch index.
            speaker_id (int): The speaker id.
        """
        speakers = self.batch_asr_states[batch_idx].get_speakers()
        for speaker_index in range(0, speaker_id + 1):
            if speaker_index not in speakers:
                self.batch_asr_states[batch_idx].add_speaker(
                    speaker_id=speaker_index,
                    asr_cache_state=self.asr_model.encoder.get_initial_cache_state(batch_size=1),
                )

    def get_speakers(self, batch_idx: int):
        """
        Get the speaker ids (int) for each instance.

        Args:
            batch_idx (int): The batch index.
        """
        return self.batch_asr_states[batch_idx].get_speakers()

    def to(self, device: torch.device):
        """
        Override the to method to move the ASR and Diar states to the device.

        Args:
            device (torch.device): The device to move the ASR and Diar states to.
        """
        for batch_idx in range(len(self.batch_asr_states)):
            self.batch_asr_states[batch_idx].to(device)
        self.diar_states.to(device)

    def update_diar_state(
        self,
        diar_pred_out_stream: torch.Tensor,
        previous_chunk_preds: torch.Tensor,
        diar_streaming_state: StreamingSortformerState,
    ):
        """
        Update the diarization state from the diarization step.
        The diarization results are updated as a form of torch.Tensor.

        Args:
            diar_pred_out_stream (torch.Tensor): The diarization prediction output stream.
            previous_chunk_preds (torch.Tensor): The previous chunk prediction output.
            diar_streaming_state (StreamingSortformerState): The diarization streaming state.
        """
        self.diar_states.diar_pred_out_stream = diar_pred_out_stream
        self.diar_states.previous_chunk_preds = previous_chunk_preds
        self.diar_states.streaming_state = diar_streaming_state

    def update_asr_state(
        self,
        batch_idx,
        speaker_id,
        cache_last_channel,
        cache_last_time,
        cache_last_channel_len,
        previous_hypotheses,
        previous_pred_out,
    ):
        """
        A function to update the ASR state with the new ASR cache state.
        This function should be called at every streaming step to update the ASR cache state.

        Args:
            batch_idx (int): The batch index.
                If parallel mode, this is the index of the potential speaker.
                If serial mode, this is the index of the session.
            speaker_id (int): The speaker id in the given session.
            -- Cache aware ASR related parameters --
            cache_last_channel (torch.Tensor)
            cache_last_time (torch.Tensor)
            cache_last_channel_len (torch.Tensor)
            previous_hypotheses (Hypothesis)
            previous_pred_out (torch.Tensor) The previous prediction output.
        """
        self.batch_asr_states[batch_idx].update_asr_state(
            speaker_id,
            cache_last_channel,
            cache_last_time,
            cache_last_channel_len,
            previous_hypotheses,
            previous_pred_out,
        )

    def get_active_speakers_info(self, active_speakers, chunk_audio, chunk_lengths):
        """
        Collect the active speaker information for the next streaming step and
        update the active speaker buffers.

        Args:
            active_speakers (List[List[int]]): The active speakers for each chunk.
            chunk_audio (torch.Tensor): The chunk audio.
            chunk_lengths (torch.Tensor): The chunk lengths.
        """
        # Reset the active speaker buffers
        self._reset_active_speaker_buffers()

        # Loop through the active speakers and update the active speaker buffers
        for batch_idx, speaker_ids in enumerate(active_speakers):
            for speaker_id in speaker_ids:
                self._active_chunk_audio.append(chunk_audio[batch_idx, :])
                self._active_chunk_lengths.append(chunk_lengths[batch_idx])
                self._active_speaker_targets.append(self.diar_states.previous_chunk_preds[batch_idx, :, speaker_id])
                inactive_speaker_ids = [i for i in range(len(speaker_ids)) if i != speaker_id]
                self._inactive_speaker_targets.append(
                    (self.diar_states.previous_chunk_preds[batch_idx, :, inactive_speaker_ids] > 0.5).sum(dim=-1) > 0
                )
                if speaker_id not in self.batch_asr_states[batch_idx].get_speakers():
                    self.add_speaker(batch_idx, speaker_id)

                self._active_previous_hypotheses.append(
                    self.batch_asr_states[batch_idx].previous_hypothesis[speaker_id]
                )
                self._active_asr_pred_out_stream.append(self.batch_asr_states[batch_idx].previous_pred_out[speaker_id])
                self._active_cache_last_channel.append(
                    self.batch_asr_states[batch_idx].cache_last_channel[:, speaker_id]
                )
                self._active_cache_last_time.append(self.batch_asr_states[batch_idx].cache_last_time[:, speaker_id])
                self._active_cache_last_channel_len.append(
                    self.batch_asr_states[batch_idx].cache_last_channel_len[speaker_id]
                )
        if len(self._active_chunk_audio) == 0:
            return None, None, None, None

        # Convert chunk audio and target info to tensors
        active_chunk_audio = torch.stack(self._active_chunk_audio)
        active_chunk_lengths = torch.stack(self._active_chunk_lengths)
        active_speaker_targets = torch.stack(self._active_speaker_targets)
        inactive_speaker_targets = torch.stack(self._inactive_speaker_targets)

        # Update active speaker attributes
        self.active_previous_hypotheses = deepcopy(self._active_previous_hypotheses)
        self.active_asr_pred_out_stream = deepcopy(self._active_asr_pred_out_stream)
        self.active_cache_last_channel = torch.stack(self._active_cache_last_channel).transpose(0, 1)
        self.active_cache_last_time = torch.stack(self._active_cache_last_time).transpose(0, 1)
        self.active_cache_last_channel_len = torch.stack(self._active_cache_last_channel_len)
        return active_chunk_audio, active_chunk_lengths, active_speaker_targets, inactive_speaker_targets

    def update_seglsts(self, offset: int):
        """
        Take the ASR states and update the seglsts.

        Args:
            offset (int): The offset of the chunk.
        """
        for asr_state in self.batch_asr_states:
            asr_state.update_sessionwise_seglsts_for_parallel(offset=offset)



# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# Copyright (c) 2024 Arc Institute. All rights reserved.
# Copyright (c) 2024 Michael Poli. All rights reserved.
# Copyright (c) 2024 Stanford University. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from copy import deepcopy
from typing import Literal, Optional

import torch
from megatron.core import parallel_state, tensor_parallel
from megatron.core.config_logger import has_config_logger_enabled, log_config_to_disk
from megatron.core.inference.contexts import BaseInferenceContext
from megatron.core.models.common.embeddings.language_model_embedding import LanguageModelEmbedding
from megatron.core.models.common.embeddings.rotary_pos_embedding import RotaryEmbedding
from megatron.core.models.common.language_module.language_module import LanguageModule
from megatron.core.packed_seq_params import PackedSeqParams
from megatron.core.process_groups_config import ProcessGroupCollection
from megatron.core.quantization.utils import get_quant_config_or_none
from megatron.core.transformer.enums import ModelType
from megatron.core.transformer.spec_utils import ModuleSpec, build_module
from megatron.core.transformer.transformer_config import TransformerConfig
from megatron.core.transformer.transformer_layer import TransformerLayer
from megatron.core.utils import WrappedTensor, deprecate_inference_params
from torch import Tensor
from torch.nn.parameter import Parameter

from nemo.collections.llm.gpt.model.megatron.hyena.hyena_config import HyenaConfig
from nemo.collections.llm.gpt.model.megatron.hyena.hyena_utils import (
    get_init_method,
    make_upper_case,
    reweighted_cross_entropy,
)


class HyenaModel(LanguageModule):
    """
    A class for the HyenaModel.
    """

    def __init__(
        self,
        transformer_config: TransformerConfig,  # Actually a hyena.HyenaConfig but avoid circular import
        hyena_stack_spec: ModuleSpec,
        vocab_size: int,
        max_sequence_length: int,
        num_groups_hyena: int,
        num_groups_hyena_medium: int,
        num_groups_hyena_short: int,
        pre_process: bool = True,
        hybrid_override_pattern: str = None,
        post_process: bool = True,
        fp16_lm_cross_entropy: bool = False,
        parallel_output: bool = True,
        post_layer_norm: bool = True,
        share_embeddings_and_output_weights: bool = True,
        position_embedding_type: Literal['learned_absolute', 'rope', 'none'] = 'rope',
        rotary_percent: float = 1.0,
        rotary_base: int = 10000,
        seq_len_interpolation_factor: Optional[float] = None,
        hyena_init_method: str = None,
        hyena_output_layer_init_method: str = None,
        remove_activation_post_first_layer: bool = True,
        add_attn_proj_bias: bool = True,
        pg_collection: Optional[ProcessGroupCollection] = None,
        vp_stage: Optional[int] = None,
    ) -> None:
        # Check if super().__init__ accepts pg_collection parameter
        super_init_signature = inspect.signature(super().__init__)
        if 'pg_collection' in super_init_signature.parameters:
            super().__init__(config=transformer_config, pg_collection=pg_collection)
        else:
            # Older version of Megatron does not initialize pg_collection yet.
            super().__init__(config=transformer_config)
            # Store pg_collection for use in submodules
            if pg_collection is None:
                pg_collection = ProcessGroupCollection.use_mpu_process_groups()
            self.pg_collection = pg_collection
            self.pp_group = pg_collection.pp

        self.transformer_config = transformer_config
        self.hyena_config = HyenaConfig()
        self.vp_stage = vp_stage

        # Override HyenaConfig fields with user provided values
        self.hyena_config.num_groups_hyena = num_groups_hyena
        self.hyena_config.num_groups_hyena_medium = num_groups_hyena_medium
        self.hyena_config.num_groups_hyena_short = num_groups_hyena_short
        if hyena_init_method:
            self.transformer_config.init_method = get_init_method(
                hyena_init_method, self.transformer_config.num_layers, self.transformer_config.hidden_size
            )
        if hyena_output_layer_init_method:
            self.transformer_config.output_layer_init_method = get_init_method(
                hyena_output_layer_init_method, self.transformer_config.num_layers, self.transformer_config.hidden_size
            )

        if has_config_logger_enabled(transformer_config):
            log_config_to_disk(transformer_config, locals(), prefix=type(self).__name__)

        self.hyena_stack_spec: ModuleSpec = hyena_stack_spec
        self.vocab_size = vocab_size
        self.max_sequence_length = max_sequence_length
        self.pre_process = pre_process
        self.hybrid_override_pattern = hybrid_override_pattern
        self.post_process = post_process
        self.fp16_lm_cross_entropy = fp16_lm_cross_entropy
        self.parallel_output = parallel_output
        self.share_embeddings_and_output_weights = share_embeddings_and_output_weights
        self.position_embedding_type = position_embedding_type
        self.post_layer_norm = post_layer_norm
        # megatron core pipelining currently depends on model type
        # TODO: remove this dependency ?
        self.model_type = ModelType.encoder_or_decoder

        if self.pre_process:
            self.embedding = LanguageModelEmbedding(
                config=self.transformer_config,
                vocab_size=self.vocab_size,
                max_sequence_length=self.max_sequence_length,
                position_embedding_type=position_embedding_type,
                tp_group=self.pg_collection.tp,
            )
        # Cache for RoPE tensors which do not change between iterations.
        self.rotary_pos_emb_cache = {}
        if self.position_embedding_type == 'rope':
            self.rotary_pos_emb = RotaryEmbedding(
                kv_channels=self.transformer_config.kv_channels,
                rotary_percent=rotary_percent,
                seq_len_interpolation_factor=seq_len_interpolation_factor,
                rotary_base=rotary_base,
                use_cpu_initialization=self.transformer_config.use_cpu_initialization,
                cp_group=self.pg_collection.cp,
            )

        self.decoder = build_module(
            hyena_stack_spec,
            self.transformer_config,
            self.hyena_config,
            hybrid_override_pattern=self.hybrid_override_pattern,
            max_sequence_length=self.max_sequence_length,
            pre_process=self.pre_process,
            post_process=self.post_process,
            post_layer_norm=self.post_layer_norm,
            pg_collection=self.pg_collection,
        )

        # In some Hyena species, the published checkpoint has identity activations after the first
        # MLP block, so we replicate this behavior in this implementation if remove_activation_post_first_layer.
        self.remove_activation_post_first_layer = remove_activation_post_first_layer
        if self.remove_activation_post_first_layer:
            if parallel_state.is_pipeline_first_stage():
                # Skip the first layer of the global model for this activation patch.
                start_idx = 1
            else:
                start_idx = 0
            mlp_no_act_config = deepcopy(self.decoder.layers[start_idx].mlp.config)
            mlp_no_act_config.activation_func = lambda x: x
            for hyena_layer in self.decoder.layers[start_idx:]:
                hyena_layer.mlp.activation_func = mlp_no_act_config.activation_func
                hyena_layer.mlp.config = mlp_no_act_config

        # In some Hyena species, the published checkpoint always has a bias in the linear projection
        # of the self-attention layers regardless of bias in other linear layers.
        self.add_attn_proj_bias = add_attn_proj_bias
        if self.add_attn_proj_bias and not self.config.add_bias_linear:
            for layer in self.decoder.layers:
                if isinstance(layer, TransformerLayer):
                    linear_proj = layer.self_attention.linear_proj
                    output_size = linear_proj.weight.shape[0]
                    linear_proj.bias = Parameter(
                        torch.empty(
                            output_size, dtype=linear_proj.config.params_dtype, device=linear_proj.weight.device
                        )
                    )
                    # Always initialize bias to zero.
                    with torch.no_grad():
                        linear_proj.bias.zero_()
                    setattr(linear_proj.bias, 'allreduce', True)
                    setattr(linear_proj, 'te_return_bias', True)
                    setattr(linear_proj, 'return_bias', True)
                    setattr(linear_proj, 'use_bias', True)
                    setattr(linear_proj.bias, 'sequence_parallel', linear_proj.config.sequence_parallel)

        # Output
        if post_process:
            if self.config.defer_embedding_wgrad_compute:
                # The embedding activation buffer preserves a reference to the input activations
                # of the final embedding projection layer GEMM. It will hold the activations for
                # all the micro-batches of a global batch for the last pipeline stage. Once we are
                # done with all the back props for all the microbatches for the last pipeline stage,
                # it will be in the pipeline flush stage. During this pipeline flush we use the
                # input activations stored in embedding activation buffer and gradient outputs
                # stored in gradient buffer to calculate the weight gradients for the embedding
                # final linear layer.
                self.embedding_activation_buffer = []
                self.grad_output_buffer = []
            else:
                self.embedding_activation_buffer = None
                self.grad_output_buffer = None
            self.output_layer = tensor_parallel.ColumnParallelLinear(
                transformer_config.hidden_size,
                self.vocab_size,
                config=transformer_config,
                init_method=transformer_config.init_method,
                bias=self.config.add_bias_output,
                skip_bias_add=False,
                gather_output=not self.parallel_output,
                skip_weight_param_allocation=self.pre_process and self.share_embeddings_and_output_weights,
                tp_group=self.pg_collection.tp,
            )
            if self.config.add_bias_output:
                self.output_layer.bias.data.zero_()

        if self.pre_process or self.post_process:
            self.setup_embeddings_and_output_layer()

        for name, module in self.named_modules():
            if hasattr(module, 'finish_init'):
                quant_config = get_quant_config_or_none(name, self.config.quant_recipe)
                module.finish_init(quant_config)

    def set_input_tensor(self, input_tensor: Tensor) -> None:
        """Sets input tensor to the model.

        See megatron.model.transformer.set_input_tensor()

        Args:
            input_tensor (Tensor): Sets the input tensor for the model.
        """
        # This is usually handled in schedules.py but some inference code still
        # gives us non-lists or None
        if not isinstance(input_tensor, list):
            input_tensor = [input_tensor]

        assert len(input_tensor) == 1, 'input_tensor should only be length 1 for gpt/bert'
        self.decoder.set_input_tensor(input_tensor[0])

    def _preprocess(
        self,
        input_ids: Tensor,
        position_ids: Tensor,
        decoder_input: Tensor | None = None,
        inference_context: BaseInferenceContext | None = None,
        packed_seq_params: PackedSeqParams | None = None,
    ):
        """Preprocesses inputs for the transformer decoder.

        Applies embeddings to input tokens, or uses `decoder_input` from a previous
        pipeline stage. Also sets up rotary positional embeddings.
        """

        # If decoder_input is provided (not None), then input_ids and position_ids are ignored.
        # Otherwise, apply embedding layer on input_ids and position_ids to get decoder_input.

        in_inference_mode = inference_context is not None and not self.training

        # Decoder embedding.
        if decoder_input is not None:
            pass
        elif self.pre_process:
            decoder_input = self.embedding(input_ids=input_ids, position_ids=position_ids)
        else:
            # intermediate stage of pipeline
            # decoder will get hidden_states from encoder.input_tensor
            decoder_input = None

        # Rotary positional embeddings (embedding is None for PP intermediate devices)
        rotary_pos_emb = None
        rotary_pos_cos = None
        rotary_pos_sin = None
        if self.position_embedding_type == 'rope' and not self.config.multi_latent_attention:
            if in_inference_mode and self.config.flash_decode:
                assert (
                    inference_context.is_static_batching()
                ), "GPTModel currently only supports static inference batching."
                # Flash decoding uses precomputed cos and sin for RoPE
                rotary_pos_cos, rotary_pos_sin = self.rotary_pos_emb_cache.setdefault(
                    inference_context.max_sequence_length,
                    self.rotary_pos_emb.get_cos_sin(inference_context.max_sequence_length),
                )
            else:
                rotary_seq_len = self.rotary_pos_emb.get_rotary_seq_len(
                    inference_context, self.decoder, decoder_input, self.config, packed_seq_params
                )
                rotary_pos_emb = self.rotary_pos_emb(
                    rotary_seq_len,
                    packed_seq=packed_seq_params is not None and packed_seq_params.qkv_format == 'thd',
                )

        if (
            in_inference_mode
            and (self.config.enable_cuda_graph or self.config.flash_decode)
            and rotary_pos_cos is not None
            and inference_context.is_static_batching()
        ):
            current_batch_size = input_ids.shape[0]
            sequence_len_offset = torch.tensor(
                [inference_context.sequence_len_offset] * current_batch_size,
                dtype=torch.int32,
                device=rotary_pos_cos.device,  # Co-locate this with the rotary tensors
            )
        else:
            sequence_len_offset = None

        # Wrap decoder_input to allow the decoder (TransformerBlock) to delete the
        # reference held by this caller function, enabling early garbage collection for
        # inference. Skip wrapping if decoder_input is logged after decoder completion.
        if in_inference_mode and not has_config_logger_enabled(self.config):
            decoder_input = WrappedTensor(decoder_input)

        return decoder_input, rotary_pos_emb, rotary_pos_cos, rotary_pos_sin, sequence_len_offset

    def forward(
        self,
        input_ids: Tensor,
        position_ids: Tensor,
        attention_mask: Tensor,
        decoder_input: Tensor = None,
        labels: Tensor = None,
        loss_mask: Tensor = None,
        inference_context: Optional[BaseInferenceContext] = None,
        packed_seq_params: PackedSeqParams = None,
        runtime_gather_output: Optional[bool] = None,
        *,
        inference_params=None,
        extra_block_kwargs=None,
    ) -> Tensor:
        """Forward pass for the HyenaModel."""
        # If decoder_input is provided (not None), then input_ids and position_ids are ignored.
        # Otherwise, apply embedding layer on input_ids and position_ids to get decoder_input.
        inference_context = deprecate_inference_params(inference_context, inference_params)
        in_inference_mode = inference_context is not None and not self.training
        if in_inference_mode:
            assert runtime_gather_output, "Inference must always gather TP logits"
        else:
            assert (
                not self.config.flash_decode
            ), "Flash decode is only supported in inference mode, but no inference_context is provided"

        decoder_input, rotary_pos_emb, rotary_pos_cos, rotary_pos_sin, sequence_len_offset = self._preprocess(
            input_ids=input_ids,
            position_ids=position_ids,
            decoder_input=decoder_input,
            inference_context=inference_context,
            packed_seq_params=packed_seq_params,
        )

        # Run decoder.
        hidden_states = self.decoder(
            hidden_states=decoder_input,
            attention_mask=attention_mask,
            inference_context=inference_context,
            rotary_pos_emb=rotary_pos_emb,
            rotary_pos_cos=rotary_pos_cos,
            rotary_pos_sin=rotary_pos_sin,
            packed_seq_params=packed_seq_params,
            sequence_len_offset=sequence_len_offset,
            **(extra_block_kwargs or {}),
        )

        if not self.post_process:
            return hidden_states

        # logits and loss
        output_weight = None
        if self.share_embeddings_and_output_weights:
            output_weight = self.shared_embedding_or_output_weight()

        if in_inference_mode and inference_context.materialize_only_last_token_logits:
            if inference_context.is_static_batching():
                hidden_states = hidden_states[-1:, :, :]
            else:
                # Reshape [B, 1, H] to [1, B, H] → extract each sample’s true last‐token hidden
                # state ([B, H]) → unsqueeze back to [1, B, H]
                # (so that the output layer, which expects S×B×H, receives only the final token)
                hidden_states = inference_context.last_token_logits(hidden_states.squeeze(1).unsqueeze(0)).unsqueeze(1)

        logits, _ = self.output_layer(hidden_states, weight=output_weight, runtime_gather_output=runtime_gather_output)
        if labels is None:
            # [s b h] => [b s h]
            return logits.transpose(0, 1).contiguous()

        labels, lowercase_mask = make_upper_case(labels)
        loss = self.compute_language_model_loss(labels, logits)
        normalize_per_batch = True if self.config.to_upper == "normalized_weighted" else False
        loss = reweighted_cross_entropy(
            loss,
            (labels, loss_mask, lowercase_mask),
            lowercase_weight=self.hyena_config.lowercase_loss_reweighting,
            normalize_per_batch=normalize_per_batch,
        )
        return loss
